"""This module implements the :py:class:`.BaseModel` class which can be
inherited from for implementing a utopya-controlled model.

Its main aim is to provide shared simulation infrastructure and does not make
further assumptions about the abstraction a model makes, like the step-wise
iteration done in :py:class:`~utopya_backend.model.step.StepwiseModel`."""

import abc
import random
import signal
import sys
import time
from typing import Optional, Union

import h5py as h5
import numpy as np

from ..logging import backend_logger as _backend_logger
from ..logging import get_level as _get_level
from ..signal import SIG_STOPCOND, SIGNAL_INFO, attach_signal_handlers
from ..tools import load_cfg_file as _load_cfg_file

# -----------------------------------------------------------------------------


class BaseModel(abc.ABC):
    """An abstract base model class that can be inherited from to implement a
    model. This class provides basic simulation infrastructure in a way that it
    couples to utopya as a frontend:

    - A shared RNG instance and a logger.
    - includes logic to evaluate ``num_steps`` and ``write_{start,every}``
    - emits monitor information to the frontend, informing the utopya frontend
      about simulation progress

    For more specific purposes, there are specialized classes that are built
    around a shared modelling paradigm like step-wise iteration of the model,
    see :py:class:`~utopya_backend.model.step.StepwiseModel`.
    """

    ATTACH_SIGNAL_HANDLERS: bool = True
    """If true, calls :py:func:`~utopya_backend.signal.attach_signal_handlers`
    to attach signal handlers for stop conditions and interrupts.

    .. hint::

        You may want to disable this in a subclass in case you cannot handle
        a case where a signal is meant to stop your simulation gracefully.
    """

    USE_SYS_EXIT: bool = True
    """If false, will not call sys.exit upon handled signal, but just return
    the error code."""

    # .. Initialization and Teardown ..........................................

    def __init__(
        self,
        *,
        cfg_file_path: str,
        _log: "logging.Logger" = None,
    ):
        """Initialize the model instance, constructing an RNG and HDF5 group
        to write the output data to.

        .. todo::

            Allow initializing from a "parent model" such that hierarchical
            nesting of models becomes possible.

        Args:
            cfg_file_path (str): The path to the config file.
            _log (logging.Logger, optional): The logger instance from which to
                create a child logger for this model. If not given, will use
                the backend logger instance.
        """
        if _log is None:
            _log = _backend_logger

        # Get the root configuration and get the instance name from there
        self._root_cfg = self._get_root_cfg(cfg_file_path, _log=_log)
        self._name = self._root_cfg["root_model_name"]

        _bases = (b.__name__ for b in type(self).__bases__)
        _log.info("Setting up model infrastructure ...")
        _log.info("  Class name:       %s", type(self).__name__)
        _log.info("  Parent class:     %s", ", ".join(_bases))
        _log.info("  Instance name:    %s", self.name)

        # Can now set up the loggers
        self._log = None
        self._setup_loggers(_log)

        # Optionally attach signal handlers for stop conditions and interrupts
        if self.ATTACH_SIGNAL_HANDLERS:
            self._signal_info = SIGNAL_INFO
            self._attach_signal_handlers()

        # Monitoring
        self.log.info("Extracting monitoring settings ...")
        self._last_emit = 0
        self._monitor_info = dict()
        self._monitor_emit_interval = self.root_cfg["monitor_emit_interval"]

        # Keep track of number of iterations
        self._n_iterations = 0

        # RNG
        # In order to ensure that a simulation is deterministic even when not
        # using the model-specific RNG instance, some singleton-like default
        # RNGs are (by default) seeded additionally. To avoid equal random
        # number sequences, they are salted.
        self._rng = self._setup_rng(
            seed=self.root_cfg["seed"],
            seed_numpy_rng=self.root_cfg.get("seed_numpy_rng", True),
            seed_system_rng=self.root_cfg.get("seed_system_rng", True),
        )

        # Create the output file
        self._h5file = self._setup_output_file()
        self._h5group = self._setup_output_group()

        # Allow subclasses to parse root config parameters, potentiaally adding
        # more attributes
        self._parse_root_cfg(**self.root_cfg)

        # Actual model initialization
        self._cfg = self.root_cfg[self.name]
        self._invoke_setup()

        # First monitoring
        self.trigger_monitor(force=True)

        # Done.
        self.log.info(
            "Fully initialized %s named '%s'.\n",
            type(self).__name__,
            self.name,
        )

    def __del__(self):
        """Takes care of tearing down the model"""
        self.log.debug("Tearing down model instance ...")

        try:
            self._h5file.close()
        except Exception as exc:
            self.log.error(
                "Closing HDF5 file failed, got %s: %s", type(exc).__name__, exc
            )

        self.log.debug("Teardown complete.")

    # .. Properties ...........................................................

    @property
    def name(self) -> str:
        """Returns the name of this model instance"""
        return self._name

    @property
    def log(self) -> "logging.Logger":
        """Returns the model's logger instance"""
        return self._log

    @property
    def rng(self) -> "numpy.random.Generator":
        """Returns the shared random number generator instance"""
        return self._rng

    @property
    def h5group(self) -> "h5py.Group":
        """The HDF5 group this model should write to"""
        return self._h5group

    @property
    def root_cfg(self) -> dict:
        """Returns the root configuration of the simulation run"""
        return self._root_cfg

    @property
    def cfg(self) -> dict:
        """Returns the model configuration, ``self.root_cfg[self.name]``"""
        return self._cfg

    @property
    def n_iterations(self) -> int:
        """Returns the number of iterations performed by this base class, i.e.
        the number of times :py:meth:`.iterate` was called.

        .. note::

            This may not correspond to potentially existing other measures that
            specialized base classes implement. For instance,
            :py:meth:`utopya_backend.model.step.StepwiseModel.time` is *not*
            the same as the number of iterations.
        """
        return self._n_iterations

    # .. Simulation control ...................................................

    def run(self) -> int:
        """Performs a simulation run for this model, calling the
        :py:meth:`.iterate` method while :py:meth:`.should_iterate` is true.
        In addition, it takes care to invoke data writing and monitoring.

        Returns:
            int: exit code, non-zero upon handled signals
        """
        self._pre_run()
        self._invoke_prolog()

        self.log.info("Commencing model run ...")

        while self.should_iterate():
            self._pre_iterate()

            self._invoke_iterate()
            self._n_iterations += 1

            # Allow to monitor simulation progress
            self.trigger_monitor()

            # Allow writing data
            if self.should_write():
                self._invoke_write_data()

            # Inform about the iteration
            self.show_iteration_info()

            # Handle signals, which may lead to a sys.exit
            if (exit_code := self._check_signals()) is not None:
                self._invoke_epilog(finished_run=False)
                self._post_run(finished_run=False)
                if self.USE_SYS_EXIT:
                    self.log.info("Now exiting (code: %d) ...", exit_code)
                    sys.exit(exit_code)
                else:
                    self.log.info("Now returning (code: %d) ...", exit_code)
                    return exit_code

            self._post_iterate()

        self._invoke_epilog(finished_run=True)
        self._post_run(finished_run=True)
        self.log.info("Simulation run finished.\n")
        return 0

    # .. Abstract methods .....................................................

    @abc.abstractmethod
    def setup(self) -> None:
        """Called upon initialization of the model"""

    @abc.abstractmethod
    def should_iterate(self) -> bool:
        """A method that determines whether :py:meth:`.iterate` should be
        called or not."""

    @abc.abstractmethod
    def iterate(self) -> None:
        """Called repeatedly until the end of the simulation, which can be
        either due to"""

    @abc.abstractmethod
    def should_write(self) -> bool:
        """A method that determines whether :py:meth:`.write_data` should be
        called after an iteration or not."""

    @abc.abstractmethod
    def write_data(self) -> None:
        """Performs data writing if :py:meth:`.should_write` returned true."""

    # .. Optionally subclassable methods ......................................

    def _parse_root_cfg(self, **_) -> None:
        """Invoked from within :py:meth:`.__init__`, parses and handles
        configuration parameters.

        .. hint:: This method can be specialized in a subclass.
        """
        pass

    def _setup_finished(self) -> None:
        """Invoked from within :py:meth:`.__init__` after the call to the
        :py:meth:`.setup` method has finished.

        .. hint:: This method can be specialized in a subclass.
        """
        pass

    def monitor(self, monitor_info: dict) -> dict:
        """Called when a monitor emission is imminent; should be used to
        update the (model-specific) monitoring information passed here as
        arguments.

        .. hint:: This method can be specialized in a subclass.
        """
        return monitor_info

    def compute_progress(self) -> float:
        """Computes the progress of the simulation run. Should return a float
        between 0 and 1 and should *always* be monotonic.

        .. hint:: This method can be specialized in a subclass.
        """
        return 0.0

    def show_iteration_info(self) -> None:
        """A method that informs about the current iteration"""
        self.log.debug("Finished iteration %d.", self.n_iterations)

    def prolog(self) -> None:
        """Invoked at the beginning of :py:meth:`.run`, before the first call
        to :py:meth:`.iterate`.

        .. hint:: This method can be specialized in a subclass.
        """
        pass

    def epilog(self, *, finished_run: bool) -> None:
        """Always invoked at the end of :py:meth:`.run`.

        This may happen either after :py:meth:`.should_iterate` returned False
        or any time before that, e.g. due to an interrupt signal or a stop
        condition. In the latter case, ``finished_run`` will be False.

        .. hint:: This method can be specialized in a subclass.
        """
        pass

    def _pre_run(self):
        """Invoked at beginning of :py:meth:`.run`"""
        pass

    def _post_run(self, *, finished_run: bool) -> None:
        """Invoked at end of :py:meth:`.run`"""
        pass

    def _pre_iterate(self):
        """Invoked at beginning of a *full* iteration"""
        pass

    def _post_iterate(self):
        """Invoked at end of a *full* iteration (including monitoring, data
        writing etc.)"""
        pass

    def _pre_monitor(self):
        """Invoked before monitor emission"""
        pass

    def _post_monitor(self):
        """Invoked after monitor emission"""
        pass

    # .. Signalling ...........................................................

    def _attach_signal_handlers(self):
        """Invoked from :py:meth:`.__init__`, attaches a signal handler for the
        stop condition signal and other interrupts.

        .. note::

            This should only be overwritten if you want or need to do your own
            signal handling.
        """
        self.log.info("Attaching signal handlers ...")
        attach_signal_handlers()

    def _check_signals(self) -> Union[None, int]:
        """Evaluates whether the iteration should stop due to an (expected)
        signal, e.g. from a stop condition or an interrupt.
        If it should stop, will return an integer, which can then be passed
        into :py:func:`sys.exit`.

        Exit codes will be ``128 + abs(signum)``, as is convention. This is
        also expected by :py:class:`~utopya.workermanager.WorkerManager` and is
        used to behave differently on a stop-condition-related signal than on
        an interrupt signal.

        Returns:
            Union[None, int]: An integer if the signal denoted that there
                should be a system exit; None otherwise.
        """
        signal_info = self._signal_info

        if not signal_info["got_signal"]:
            return None

        # Received a signal
        signum = signal_info["signum"]
        if signum == SIG_STOPCOND:
            self.log.warning("A stop condition was fulfilled.")

        elif signum in (signal.SIGINT, signal.SIGTERM):
            self.log.warning("Was told to stop.")

        else:
            self.log.warning(
                "Got an unexpected signal: %d. Stopping ...", signum
            )

        return 128 + abs(signum)

    # .. Monitoring ...........................................................

    def _monitor_should_emit(self, *, t: float = None) -> bool:
        """Evaluates whether the monitor should emit. This method will only
        return True once a monitor emit interval has passed since the last
        time the monitor was emitted.

        Args:
            t (None, optional): If given, uses this time, otherwise calls
                :py:func:`time.time`.

        Returns:
            bool: Whether to emit or not.
        """
        t = t if t is not None else time.time()

        if t > self._last_emit + self._monitor_emit_interval:
            return True
        return False

    def _emit_monitor(self):
        """Actually emits the monitoring information using :py:func:`print`."""
        # TODO Consider using YAML for creating the monitor string
        def parse_val(v) -> str:
            if isinstance(v, float):
                return f"{v:6g}"
            return repr(v)

        progress = str(self.compute_progress())
        monitor_info = ", ".join(
            f"{k}: {parse_val(v)}" for k, v in self._monitor_info.items()
        )

        # Now emit ...
        # fmt: off
        print(
            "!!map { "
            + "progress: " + progress + ", "
            + "n_iter: " + str(self.n_iterations) + ", "
            + self.name + ": {" + monitor_info + "}"
            + "}",
            flush=True,
        )
        # fmt: on

    def trigger_monitor(self, *, force: bool = False):
        """Invokes the monitoring procedure:

        #. Checks whether :py:meth:`._monitor_should_emit`.
        #. If so, calls :py:meth:`.monitor` to update monitoring information.
        #. Then calls :py:meth:`._emit_monitor` to emit that information.

        If ``force`` is given, will always emit.

        .. hint::

            This method should not be subclassed, but it can be invoked from
            within the subclass at any desired point.
        """
        t = time.time()
        if force or self._monitor_should_emit(t=t):
            self._pre_monitor()

            self._monitor_info = self.monitor(self._monitor_info)
            self._emit_monitor()
            self._last_emit = t

            self._post_monitor()

    # .. Other helpers ........................................................

    def _get_root_cfg(
        self, cfg_file_path: str, *, _log: "logging.Logger"
    ) -> dict:
        """Retrieves the root configuration for this simulation run by loading
        it from the given file path.
        """
        _log.info("Loading configuration file ...\n  %s", cfg_file_path)
        return _load_cfg_file(cfg_file_path)

    def _setup_loggers(self, _log: "logging.Logger"):
        """Sets up the model logger and configures the backend logger according
        to the ``log_levels`` entry set in the root configuration.

        .. todo::

            Allow setting the logging *format* as well.

        Args:
            _log (logging.Logger): The logger to initialize the model logger
                from, typically the
        """
        _log.info("Setting up loggers ...")

        # TODO Allow changing log formatters globally

        # Model logger and its level
        self._log = _log.getChild(self.name)

        log_level = self.root_cfg["log_levels"]["model"]
        self._log.setLevel(_get_level(log_level))
        self.log.info("  Model logger initialized with '%s' level.", log_level)

        # May want to adjust the backend logger
        backend_log_level = self.root_cfg["log_levels"].get("backend")
        if backend_log_level is not None:
            _backend_logger.setLevel(_get_level(backend_log_level))
            self.log.info(
                "  Set backend logger's level to '%s'.", backend_log_level
            )

    def _setup_rng(
        self,
        *,
        seed: int,
        seed_numpy_rng: Optional[Union[bool, int]] = None,
        seed_system_rng: Optional[Union[bool, int]] = None,
        **rng_kwargs,
    ) -> "numpy.random.Generator":
        """Sets up the shared RNG.

        .. note::

            If also seeding the other RNGs, make sure to use different seeds
            for them, such that random number sequences are ensured to be
            different even if the underlying generator may be the same.

        Args:
            seed (int): The seed for the new, model-specific RNG, constructed
                via :py:func:`numpy.random.default_rng`
            seed_numpy_rng (Optional[Union[bool, int]], optional): If not
                False or None, will also seed numpy's singleton (i.e.
                *default*) RNG by calling :py:func:`numpy.random.seed`.
                If True, will use ``seed + 1`` for that.
            seed_system_rng (Optional[Union[bool, int]], optional): If not
                False or None, will also seed the system's default RNG by
                calling :py:func:`random.seed`.
                If True, will use ``seed + 2`` for that.
            **rng_kwargs: Passed on to :py:func:`numpy.random.default_rng`
        """
        self.log.info("Creating shared RNG (seed: %s) ...", seed)
        rng = np.random.default_rng(seed, **rng_kwargs)

        if seed_numpy_rng not in (None, False):
            if seed_numpy_rng is True:
                seed_numpy_rng = seed + 1
            np.random.seed(seed_numpy_rng)
            self.log.debug(
                "  Default numpy RNG seeded with:  %s", seed_numpy_rng
            )

        if seed_system_rng not in (None, False):
            if seed_system_rng is True:
                seed_system_rng = seed + 2
            random.seed(seed_system_rng)
            self.log.debug(
                "  Default system RNG seeded with: %s", seed_system_rng
            )

        return rng

    def _setup_output_file(self) -> "h5py.File":
        """Creates the output file for this model; by default, it is a HDF5
        file that is managed by a :py:class:`h5py.File` object.

        .. note::

            This method can be subclassed to implement different output file
            formats. In that case, consider not using the ``_h5file`` and
            ``_h5group`` attributes but something else.
        """
        self.log.info(
            "Creating HDF5 output file at:\n  %s\n",
            self.root_cfg["output_path"],
        )
        return h5.File(self.root_cfg["output_path"], mode="x")

    def _setup_output_group(self, h5file: "h5py.File" = None) -> "h5py.Group":
        """Creates the group that this model's output is written to"""
        if h5file is None:
            h5file = self._h5file
        return h5file.create_group(self.name)

    # .. Invocation wrappers ..................................................

    def _invoke_iterate(self):
        self.iterate()

    def _invoke_write_data(self):
        self.write_data()

    def _invoke_setup(self):
        self.log.info("Invoking model setup ...")
        self.setup(**self.cfg)
        self.log.info("Model setup finished.\n")
        self._setup_finished()

    def _invoke_prolog(self):
        """Helps invoking the :py:meth:`.prolog`"""
        self.log.debug("Invoking prolog ...")
        self.prolog()
        self.log.debug("Prolog finished.\n")

    def _invoke_epilog(self, **kwargs):
        """Helps invoking the :py:meth:`.epilog`"""
        self.log.debug("Invoking epilog ...")
        self.epilog(**kwargs)
        self.log.debug("Epilog finished.\n")



"""utopya-specific exception types"""


class UtopyaException(BaseException):
    """Base class for utopya-specific exceptions"""


# -- Parameter validation -----------------------------------------------------


class ValidationError(UtopyaException, ValueError):
    """Raised upon failure to validate a parameter"""


# -- WorkerManager ------------------------------------------------------------


class WorkerManagerError(UtopyaException):
    """The base exception class for WorkerManager errors"""


class WorkerManagerTotalTimeout(WorkerManagerError):
    """Raised when a total timeout occurred"""


class WorkerTaskError(WorkerManagerError):
    """Raised when there was an error in a WorkerTask"""


class WorkerTaskNonZeroExit(WorkerTaskError):
    """Can be raised when a WorkerTask exited with a non-zero exit code."""

    def __init__(self, task: "utopya.task.WorkerTask", *args, **kwargs):
        """Initialize an error handling non-zero exit codes from workers"""
        self.task = task
        super().__init__(*args, **kwargs)

    def __str__(self) -> str:
        """Returns information on the error"""
        from ._signal import SIGMAP

        signals = [
            signal
            for signal, signum in SIGMAP.items()
            if signum == abs(self.task.worker_status)
        ]

        return (
            f"Task '{self.task.name}' exited with non-zero exit "
            f"status: {self.task.worker_status}.\nThis may originate from "
            f"the following signals:  {', '.join(signals)}.\n"
            "Googling these might help with identifying the error. "
            "Also, inspect the log and the log file for further error "
            "messages. To increase verbosity, run in debug mode, e.g. by "
            "passing the --debug flag to the CLI."
        )


class WorkerTaskStopConditionFulfilled(WorkerTaskNonZeroExit):
    """An exception that is raised when a worker-specific stop condition was
    fulfilled. This allows being handled separately to other non-zero exits.
    """


# -----------------------------------------------------------------------------


class YAMLRegistryError(UtopyaException, ValueError):
    """Base class for errors in YAMLRegistry"""


class EntryExistsError(YAMLRegistryError):
    """Raised if an entry already exists"""


class MissingEntryError(YAMLRegistryError):
    """Raised if an entry is missing"""


class MissingRegistryError(YAMLRegistryError):
    """Raised if a registry is missing"""


class EntryValidationError(YAMLRegistryError):
    """Raised upon failed validation of a registry entry"""


class SchemaValidationError(YAMLRegistryError):
    """If schema validation failed"""


# -----------------------------------------------------------------------------


class ModelRegistryError(UtopyaException, ValueError):
    """Raised on errors with model registry"""


class MissingModelError(ModelRegistryError):
    """Raised when a model is missing"""


class BundleExistsError(ModelRegistryError):
    """Raised when a bundle that compared equal already exists"""


class MissingBundleError(ModelRegistryError):
    """Raised when a bundle is missing"""


class BundleValidationError(ModelRegistryError):
    """Raised when the result of validating the existence of a bundle fails"""


class ProjectRegistryError(UtopyaException, ValueError):
    """Raised on errors with project registry"""


class MissingProjectError(ProjectRegistryError):
    """Raised on a missing project"""


class ProjectExistsError(ProjectRegistryError):
    """Raised if a project or project file of that name already exists"""



"""A generic, DAG-supporting time series plot"""

import xarray as xr

from . import PlotHelper, is_plot_func

# -----------------------------------------------------------------------------


@is_plot_func(
    use_dag=True,
    required_dag_tags=("data",),
    helper_defaults=dict(set_limits=dict(x=["min", "max"])),
)
def time_series(
    *,
    data: dict,
    hlpr: PlotHelper,
    x: str = "time",
    label_fstr: str = "{:.2g}",
    **plot_kwargs,
):
    """This is a generic plotting function that plots one or multiple time
    series from the ``data`` tag that is selected via the DAG framework.

    The data needs to be an xarray object. If y is an xr.DataArray, it is
    assumed to be one- or two-dimensional.
    If is an :py:class:`xarray.Dataset`, all data variables are plotted and
    their name is used as the label.

    For the x axis values, the corresponding ``time`` coordinates are used;
    these need to be part of the dataset!

    .. note::

        For a more generic plot, see dantro's
        :py:func:`~dantro.plot.funcs.generic.facet_grid`, which is available
        under ``.plot.facet_grid`` in utopya.

    Args:
        data (dict): The data selected by the DAG framework
        hlpr (PlotHelper): The plot helper
        x (str, optional): Name of the coordinate dimension to put on the
            x-axis, typically (and by default) ``time``.
        label_fstr (str, optional): Formatting to use for label in case of
            the data being an :py:class:`xarray.DataArray`.
        **plot_kwargs: Passed on ot :py:func:`matplotlib.pyplot.plot`.
    """
    d = data["data"]

    # If this is an xr.DataArray, it may be one or two-dimensional
    if isinstance(d, xr.Dataset):
        # Simply plot all data variables as individual lines
        for dvar, line in d.data_vars.items():
            hlpr.ax.plot(line.coords[x], line, label=dvar, **plot_kwargs)

        hlpr.invoke_helper(
            "set_labels", x=x.capitalize(), mark_disabled_after_use=False
        )

    elif isinstance(d, xr.DataArray):
        # Also allow two-dimensional arrays
        if d.ndim == 1:
            hlpr.ax.plot(d.coords[x], d, **plot_kwargs)

        elif d.ndim == 2:
            loop_dim = [dim for dim in d.dims if dim != x][0]

            for c in d.coords[loop_dim]:
                line = d.sel({loop_dim: c})
                hlpr.ax.plot(
                    line.coords[x],
                    line,
                    label=label_fstr.format(c.item()),
                    **plot_kwargs,
                )

            # Provide a default title to the legend: name of the loop dimension
            hlpr.invoke_helper(
                "set_legend",
                title=f"${loop_dim}$ coordinate",
                mark_disabled_after_use=False,
            )
            hlpr.invoke_helper(
                "set_labels", x=x.capitalize(), mark_disabled_after_use=False
            )

        else:
            raise ValueError(
                f"Array data needs to be 1D or 2D, but was {d.ndim}D!\n"
                f"Given data:\n{d}"
            )

    else:
        raise TypeError(f"Expected xr.Dataset or xr.DataArray, got {type(d)}")



"""Tests the utopya models <...> subcommands of the CLI"""

import os
import pickle
import traceback
from shutil import rmtree

import pytest

import utopya

from .._fixtures import *
from . import invoke_cli

# -----------------------------------------------------------------------------


def test_list(with_test_models):
    """Tests utopya models ls"""
    # Lists models as expected
    res = invoke_cli(("models", "ls"))
    assert "model registry" in res.output
    assert DUMMY_MODEL in res.output

    # Also shows the number of bundles available in "long" mode
    res = invoke_cli(("models", "ls", "--long"))
    assert "bundle" in res.output
    assert "Default bundles are marked" in res.output
    assert "(*)" in res.output

    assert res.output == invoke_cli(("models", "ls", "-l")).output


def test_register_single(registry):
    """Tests utopya models register single"""
    reg_args = (
        "models",
        "register",
        "single",
        DUMMY_MODEL,
        "-e",
        DUMMY_EXECUTABLE,
    )
    reg_args += ("--label", TEST_LABEL)  # already exists
    reg_args += ("--source-dir", "./")

    res = invoke_cli(reg_args)
    print(res.output)
    assert res.exit_code != 0
    assert "Registration failed!" in res.output
    assert "Bundle validation failed" in res.output


def test_register_from_list(registry):
    """Tests utopya models register single

    NOTE The fixture already performs some of the tests
    """
    reg_args = (
        "models",
        "register",
        "from-list",
        DUMMY_MODEL,
        "--executables",
        DUMMY_EXECUTABLE,
    )
    res = invoke_cli(reg_args + ("--label", "some_new_label"))
    print(res.output)
    if res.exception:
        traceback.print_tb(res.exception.__traceback__)
    assert res.exit_code == 0
    assert "Model registration succeeded" in res.output

    # Can also use format strings
    reg_args = (
        "models",
        "register",
        "from-list",
        DUMMY_MODEL,
        "--executable-fstr",
        "{model_name:}/{model_name:}.py",
        "--source-dir-fstr",
        "{model_name:}/",
        "--base-executable-dir",
        os.path.join(DEMO_DIR, "models"),
        "--base-source-dir",
        os.path.join(DEMO_DIR, "models"),
    )
    res = invoke_cli(reg_args + ("--label", "yet_another_label"))
    print(res.output)
    assert res.exit_code == 0
    assert "Model registration succeeded" in res.output

    # Mutually exclusive
    res = invoke_cli(
        (
            "models",
            "register",
            "from-list",
            DUMMY_MODEL,
            "--executable-fstr",
            "{model_name:}/{model_name:}.py",
            "--source-dir-fstr",
            "{model_name:}/",
            "--source-dirs",
            "/foo/bar",
        )
    )
    print(res.output)
    assert res.exit_code != 0
    assert "mutually exclusive" in res.output

    # Superfluous arguments
    res = invoke_cli(reg_args + ("--executables", "'foo;bar'"))
    print(res.output)
    assert res.exit_code != 0
    assert "mutually exclusive" in res.output

    # Missing arguments
    reg_args = (
        "models",
        "register",
        "from-list",
        DUMMY_MODEL,
    )
    res = invoke_cli(reg_args)
    print(res.output)
    assert res.exit_code != 0
    assert "Missing argument --executables or" in res.output

    # Length mismatch
    reg_args = (
        "models",
        "register",
        "from-list",
        f"'{DUMMY_MODEL};{DUMMY_MODEL}'",
        "--executables",
        DUMMY_EXECUTABLE,
    )
    res = invoke_cli(reg_args)
    print(res.output)
    assert res.exit_code != 0
    assert "Mismatch of sequence lengths" in res.output


def test_register_from_manifest(registry):
    """Tests utopya models register from-manifest"""

    # Workaround for side effects of DUMMY_MODEL persisting beyond a single
    # invocation, thus leading to test failures for repeated test calls.
    res = invoke_cli(("models", "rm", DUMMY_MODEL, "--all"), input="y\n")

    # Get path to directory
    DUMMY_INFO = os.path.join(
        DEMO_DIR, "models", DUMMY_MODEL, f"{DUMMY_MODEL}_info.yml"
    )
    reg_args = ("models", "register", "from-manifest", DUMMY_INFO)
    reg_args += ("--model-name", DUMMY_MODEL)

    res = invoke_cli(reg_args)
    print(res.output)
    assert res.exit_code == 0
    assert "from_manifest_file" in res.output

    # With custom label
    extd_reg_args = reg_args + ("--label", "custom_label")
    res = invoke_cli(extd_reg_args)
    print(res.output)
    assert res.exit_code == 0
    assert "from_manifest_file" not in res.output
    assert "custom_label" in res.output
    assert "custom_label" in registry[DUMMY_MODEL]

    # With custom model name
    extd_reg_args += ("--model-name", "MyCustomModelName")
    res = invoke_cli(extd_reg_args)
    print(res.output)
    assert len(registry["MyCustomModelName"]) == 1
    assert "custom_label" in registry["MyCustomModelName"]
    registry.remove_entry("MyCustomModelName")

    # Custom model name fails with more than one manifest file
    res = invoke_cli(extd_reg_args + (DUMMY_INFO,))
    print(res.output)
    assert res.exit_code != 0
    assert "can only be specified if only a single manifest file" in res.output


def test_remove(registry):
    """Tests utopya models rm

    This is based on the registry fixture, which adds a bunch of entries to
    the model registry which can then be removed here.
    """
    # Add some additional entries (to have something to remove)
    labels = (
        f"{TEST_LABEL}_foo",
        f"{TEST_LABEL}_bar",
        f"{TEST_LABEL}_spam",
    )
    for label in labels:
        reg_args = (
            "models",
            "register",
            "single",
            DUMMY_MODEL,
            "-e",
            DUMMY_EXECUTABLE,
        )
        reg_args += ("--label", label)
        reg_args += ("--source-dir", "./")

        res = invoke_cli(reg_args)
        print(res.output)
        assert res.exit_code == 0

    # Remove a single label
    res = invoke_cli(("models", "rm", DUMMY_MODEL, "--label", labels[0]))
    assert res.exit_code == 0
    assert TEST_LABEL in registry[DUMMY_MODEL]
    assert labels[0] not in registry[DUMMY_MODEL]
    assert labels[1] in registry[DUMMY_MODEL]
    assert labels[2] in registry[DUMMY_MODEL]

    # ... with prompt for bundle name
    res = invoke_cli(("models", "rm", DUMMY_MODEL), input=f"{labels[1]}\n")
    assert res.exit_code == 0
    assert TEST_LABEL in registry[DUMMY_MODEL]
    assert labels[0] not in registry[DUMMY_MODEL]
    assert labels[1] not in registry[DUMMY_MODEL]
    assert labels[2] in registry[DUMMY_MODEL]

    # Now remove it completely
    res = invoke_cli(("models", "rm", DUMMY_MODEL, "--all"), input="N\n")
    print(res.output)
    assert res.exit_code == 0
    assert "Not removing anything" in res.output

    res = invoke_cli(("models", "rm", DUMMY_MODEL, "--all"), input="y\n")
    print(res.output)
    assert res.exit_code == 0
    assert DUMMY_MODEL not in registry


def test_set_default(registry):
    """Tests utopya models set-default"""
    assert registry[DUMMY_MODEL].default_label == TEST_LABEL

    # Unset default and re-set via CLI
    registry[DUMMY_MODEL].default_label = None
    res = invoke_cli(("models", "set-default", DUMMY_MODEL, TEST_LABEL))
    print(res.output)
    assert res.exit_code == 0
    assert TEST_LABEL in res.output
    assert registry[DUMMY_MODEL].default_label == TEST_LABEL

    # Fails for invalid label name, keeping the old default
    assert "invalid_label" not in registry[DUMMY_MODEL]
    res = invoke_cli(("models", "set-default", DUMMY_MODEL, "invalid_label"))
    print(res.output)
    assert res.exit_code != 0
    assert "invalid_label" in res.output
    assert registry[DUMMY_MODEL].default_label == TEST_LABEL


def test_edit(monkeypatch, with_test_models):
    """Tests utopya models edit"""
    res = invoke_cli(("models", "edit", DUMMY_MODEL), input="y\n")
    print(res.output)
    assert res.exit_code == 1
    assert "Editing model registry file failed!" in res.output

    monkeypatch.setenv("EDITOR", "echo")  # this will always work
    res = invoke_cli(("models", "edit", DUMMY_MODEL), input="y\n")
    print(res.output)
    assert res.exit_code == 0
    assert "Successfully edited registry file" in res.output

    # Not continuing
    res = invoke_cli(("models", "edit", DUMMY_MODEL), input="N\n")
    print(res.output)
    assert res.exit_code == 0
    assert "Not opening" in res.output


def test_copy(registry):
    """Tests utopya models copy"""
    cmd = ("models", "copy", ADVANCED_MODEL)
    shared_args = ("--dry-run", "--yes")
    COPIED_MODEL_NAME = f"_CopyTest_{ADVANCED_MODEL}"

    # Simple invocation with all parameters given
    args = (
        "--new-name",
        COPIED_MODEL_NAME,
        "--target-project",
        TEST_PROJECT_NAME,
    )
    res = invoke_cli(cmd + args + shared_args)
    assert res.exit_code == 0

    # Can disable postprocessing
    res = invoke_cli(cmd + args + shared_args + ("--no-pp",))
    assert res.exit_code == 0
    assert "Post-processing routines were disabled." in res.output

    # And may not be getting a prompt
    res = invoke_cli(cmd + args + ("--dry-run",))
    assert res.exit_code == 0

    # Can pass file extensions to skip, which may also lead to empty file maps
    res = invoke_cli(cmd + args + shared_args + ("--skip-exts", "pyc py .yml"))
    print(res.output)
    assert res.exit_code == 0
    assert ".pyc .py .yml" in res.output
    assert "check the file extensions" in res.output
    assert "No files found to copy!" in res.output  # ... as all are ignored

    # Informs about py_tests_dir or py_plots_dir not being defined
    res = invoke_cli(("models", "copy", DUMMY_MODEL) + args + shared_args)
    print(res.output, res.exception)
    assert res.exit_code == 0
    assert "do not define a 'py_tests_dir'" in res.output
    assert "do not define a 'py_plots_dir'" in res.output

    # .. Actually copy ........................................................
    model = registry[ADVANCED_MODEL].item()
    prj = model.project

    try:
        res = invoke_cli(cmd + args + ("--yes",))
        print(res.output)
        assert "Copying failed" not in res.output
        assert res.exit_code == 0

        # And do it again, which should trigger writing errors
        res = invoke_cli(cmd + args + ("--yes",))
        print(res.output)
        assert "Copying failed with FileExistsError" in res.output
        assert res.exit_code == 0

        # Copying error will also raise
        res = invoke_cli(cmd + args + ("--yes", "--debug"))
        print(res.output)
        assert res.exit_code != 0

        # File name replacements were carried out
        model_dir = os.path.join(prj.paths["models_dir"], COPIED_MODEL_NAME)
        cfg_file = os.path.join(model_dir, f"{COPIED_MODEL_NAME}_cfg.yml")
        assert os.path.isfile(cfg_file)

        # Replacements took place
        info_file = os.path.join(model_dir, f"{COPIED_MODEL_NAME}_info.yml")
        assert os.path.isfile(info_file)
        with open(info_file) as f:
            info_file = f.read()

        print(info_file)
        assert f"model_name: {COPIED_MODEL_NAME}" in info_file

        # .. Postprocessing . . . . . . . . . . . . . . . . . . . . . . . . . .
        cml_root = os.path.join(prj.paths["models_dir"], "CMakeLists.txt")
        cml_model = os.path.join(
            prj.paths["models_dir"], ADVANCED_MODEL, "CMakeLists.txt"
        )

        # With a CMakeLists.txt file in the file map, should trigger automatic
        # postprocessing ... regardless of content
        with open(cml_model, "w") as f:
            f.write("")

        res = invoke_cli(cmd + args + shared_args)
        print(res.output)
        assert "No CMakeLists.txt file found in expected loc" in res.output
        assert res.exit_code == 0

        # With the root file existing, extending it is attempted, but the
        # add_subdirectory command is missing.
        with open(cml_root, "w") as f:
            f.write("\n")

        res = invoke_cli(cmd + args + shared_args)
        print(res.output)
        assert "inserting at the end" in res.output
        assert "preview of how the new" in res.output
        assert res.exit_code == 0

        # Now with some add_subdirectory commands existing
        with open(cml_root, "a") as f:
            f.write("\n")
            f.write("# Some comment\n")
            f.write(f"add_subdirectory(__i_should_be_first__)\n")
            f.write(f"add_subdirectory({ADVANCED_MODEL})\n")
            f.write(f"add_subdirectory({DUMMY_MODEL})\n")
            f.write("\n")
            f.write("# More content here\n")
            f.write("\n")
            f.write("# End of file\n")

        res = invoke_cli(cmd + args + ("--yes",))
        print(res.output)
        assert res.exit_code == 0

        # ... the location is correct
        with open(cml_root, "r") as f:
            print("Written file:")
            f = f.read()
            print(f)
            assert f.find(COPIED_MODEL_NAME) > f.find("__i_should_be_first__")
            assert f.find(COPIED_MODEL_NAME) < f.find(ADVANCED_MODEL)
            assert f.find(COPIED_MODEL_NAME) < f.find(DUMMY_MODEL)
            assert f.find(COPIED_MODEL_NAME) < f.find("End of file")

        # Insert behind the last add_subdirectory command
        with open(cml_root, "w") as f:
            f.write("\n")
            f.write("# Some comment\n")
            f.write(f"add_subdirectory(__i_should_be_first__)\n")
            f.write("\n")
            f.write("# End of file\n")

        res = invoke_cli(cmd + args + ("--yes",))
        print(res.output)
        assert res.exit_code == 0

        with open(cml_root, "r") as f:
            print("Written file:")
            f = f.read()
            print(f)
            assert f.find(COPIED_MODEL_NAME) > f.find("__i_should_be_first__")
            assert f.find(COPIED_MODEL_NAME) < f.find("End of file")

    finally:
        # Remove artifacts
        rmtree(os.path.join(prj.paths["models_dir"], COPIED_MODEL_NAME))
        rmtree(os.path.join(prj.paths["py_tests_dir"], COPIED_MODEL_NAME))
        rmtree(os.path.join(prj.paths["py_plots_dir"], COPIED_MODEL_NAME))
        os.remove(cml_model)
        os.remove(cml_root)

    # .. Errors ...............................................................
    # Provoke a reading error by adding a binary file
    bad_file = os.path.join(prj.paths["models_dir"], ADVANCED_MODEL, "so_bad")
    try:
        with open(bad_file, mode="wb") as f:
            pickle.dump("some object", f)

        res = invoke_cli(cmd + args + shared_args)
        print(res.output)
        assert "Reading failed" in res.output
        assert res.exit_code == 0

        res = invoke_cli(cmd + args + shared_args + ("--debug",))
        print(res.output)
        assert "Reading failed" in res.output
        assert res.exit_code != 0

    finally:
        os.remove(bad_file)

    # Existing model name throws an error
    args = (
        "--new-name",
        ADVANCED_MODEL,  # already exists, of course
        "--target-project",
        TEST_PROJECT_NAME,
    )
    res = invoke_cli(cmd + args + shared_args)
    assert res.exit_code == 1



"""Tests the utopya_backend.tools module"""

import os

import pytest
from pkg_resources import resource_filename

import utopya_backend.tools as t

# -----------------------------------------------------------------------------


def test_load_cfg_file(tmpdir):
    """Tests the generic config file loading function"""
    # Attempts to load files (which are not there, of course)
    with pytest.raises(FileNotFoundError):
        t.load_cfg_file(tmpdir.join("some_file.yml"))
    with pytest.raises(FileNotFoundError):
        t.load_cfg_file(tmpdir.join("some_file.yaml"))
    with pytest.raises(FileNotFoundError):
        t.load_cfg_file(tmpdir.join("some_file.yAmL"))

    # Can also explicitly specify a loader
    with pytest.raises(FileNotFoundError):
        t.load_cfg_file(tmpdir.join("some_file.foobar"), loader="yaml")

    # Otherwise will get an earlier error due to a unsupported load format
    with pytest.raises(ValueError, match="Unsupported loader"):
        t.load_cfg_file(tmpdir.join("some_file.foobar"))

    # Can also load an actual file
    fpath = resource_filename("utopya", "cfg/base_cfg.yml")
    cfg = t.load_cfg_file(fpath)
    assert "parameter_space" in cfg

    cfg = t.load_cfg_file(fpath, loader="yAmL")
    assert "parameter_space" in cfg


def test_import_package_from_dir(tmpdir):
    """Tests the path-based import function"""
    # Import the backend testing module right here
    mod_path = os.path.dirname(__file__)
    mod = t.import_package_from_dir(mod_path)
    assert mod.__file__.endswith("backend/__init__.py")

    # Can also specify a module string:
    mod2 = t.import_package_from_dir(mod_path, mod_str="backend")
    assert mod2 == mod

    # Can also import from higher-up, without needing to adjust the directory
    mod3 = t.import_package_from_dir(mod_path, mod_str="tests.backend")
    assert mod3.__file__.endswith("backend/__init__.py")

    # Is robust against trailing slash
    mod4 = t.import_package_from_dir(mod_path + "/", mod_str="tests.backend")
    assert mod4.__file__.endswith("backend/__init__.py")

    # Path needs to be absolute
    with pytest.raises(ValueError, match="Need an absolute path"):
        t.import_package_from_dir("some/relative/path")

    # Path needs to point to an existing *directory*
    with pytest.raises(FileNotFoundError, match="existing directory"):
        t.import_package_from_dir(__file__)

    # Path needs to point to an *existing* directory
    with pytest.raises(FileNotFoundError, match="existing directory"):
        t.import_package_from_dir("~/some/imaginary/directory")

    # Mock failing import
    mod_path = tmpdir.join("my_test_module")
    os.makedirs(mod_path)
    with open(mod_path.join("__init__.py"), "x") as f:
        f.write("raise\n")

    with pytest.raises(ImportError, match="Failed importing module 'my_test_"):
        t.import_package_from_dir(mod_path)



"""This module provides plotting functions to visualize the attractive set of
a dynamical system.

.. todo::

    Migrate these to the more generic DAG-based interface.
"""

import copy
import itertools
import logging
from typing import Callable, Dict, Sequence, Tuple, Union

import matplotlib as mpl
import numpy as np
import xarray as xr
from matplotlib.collections import PatchCollection
from matplotlib.patches import Circle, Rectangle
from scipy.signal import find_peaks

# FIXME use local import
import utopya.eval.plots._attractor as utdp

from ...tools import recursive_update
from .. import DataManager, UniverseGroup
from . import MultiversePlotCreator, PlotHelper, is_plot_func
from ._mpl import HandlerEllipse
from ._utils import calc_pxmap_rectangles

log = logging.getLogger(__name__)

# -----------------------------------------------------------------------------


@is_plot_func(creator="multiverse")
def bifurcation_diagram(
    dm: DataManager,
    *,
    hlpr: PlotHelper,
    mv_data: xr.Dataset,
    dim: str = None,
    dims: Tuple[str, str] = None,
    analysis_steps: Sequence[Union[str, Tuple[str, str]]],
    custom_analysis_funcs: Dict[str, Callable] = None,
    analysis_kwargs: dict = None,
    visualization_kwargs: dict = None,
    to_plot: dict = None,
    **kwargs,
) -> None:
    """Plots a bifurcation diagram for one or two parameter dimensions
       (arguments ``dim`` or ``dims``).

    Args:
        dm (DataManager): The data manager from which to retrieve the data
        hlpr (PlotHelper): The PlotHelper that instantiates the figure and
            takes care of plot aesthetics (labels, title, ...) and saving
        mv_data (xarray.Dataset): The extracted multidimensional dataset
        dim (str, optional): The required parameter dimension of the 1d
            bifurcation diagram.
        dims (str, optional): The required parameter dimensions (x, y) of the
            2d-bifurcation diagram.
        analysis_steps (Sequence): The analysis steps that are to be made
            until one is conclusive. Applied per universe.

            - If seq of str: The str will also be used as attractor key for
                plotting if the test is conclusive.
            - If seq of Tuple(str, str): The first str defines the attractor
                key for plotting, the second str is a key within
                custom_analysis_funcs.

            Default analysis_funcs are:

            - endpoint: utopya.dataprocessing.find_endpoint
            - fixpoint: utopya.dataprocessing.find_fixpoint
            - multistability: utdp.find_multistability
            - oscillation: utdp.find_oscillation
            - scatter: resolve_scatter
        custom_analysis_funcs (dict): A collection of custom analysis functions
            that will overwrite the default analysis funcs (recursive update).
        analysis_kwargs (dict, optional): The entries need to match the
            analysis_steps. The subentry (dict) is passed on to the analysis
            function.
        visualization_kwargs (dict, optional): The entries need to match the
            analysis_steps. The subentry (dict) is used to configure a
            rectangle to visualize the conclusive analysis step. Is passed to
            matplotlib.patches.rectangle. xy, width, height, and angle are
            ignored and set automatically. Required in 2d bifurcation diagram.
        to_plot (dict, optional): The configuration for the data to plot. The
            entries of this key need to match the data_vars selected in mv_data.
            It is used to visualize the state of the attractor additionally to
            the visualization kwargs. Only for 1d-bifurcation diagram.
            sub_keys:

            - ``label`` (str, optional): label in plot
            - ``plot_kwargs`` (dict, optional): passed to scatter for every
                universe

                - color (str, recommended): unique color for every
                    data_variable accross universes
        **kwargs: Collection of optional dicts passed to different functions

            - plot_coords_kwargs (dict): Passed to ax.scatter to mark the
                universe's center in the bifurcation diagram
            - rectangle_map_kwargs (dict): Passed to
                utopya.eval.plots._utils.calc_pxmap_rectangles
            - legend_kwargs (dict): Passed to ax.legend
    """

    def resolve_analysis_steps(
        analysis_steps: Sequence[Union[str, Tuple[str, str]]]
    ) -> Sequence[Tuple[str, str]]:
        """Resolve instance of str to Tuple[str, str] in sequence

        Args:
            analysis_steps (Sequence[Union[str, Tuple[str, str]]]): The
                original sequence

        Returns:
            analysis_steps (Sequence[Tuple[str, str]]): The sequence of
                attractor_key and analysis_func pairs.
        """
        for i, analysis_step in enumerate(analysis_steps):
            # get key and func for the analysis step
            if isinstance(analysis_step, str):
                analysis_steps[i] = [analysis_step, analysis_step]

        return analysis_steps

    def resolve_to_plot_kwargs(to_plot: dict) -> dict:
        """Resolves the to_plot dict, e.g. adding labels if not explicitly
        specified.

        Args:
            to_plot (dict): The to_plot dict to parse

        Returns:
            dara_vars_plot_kwargs (dict): A dict with the 'plot_kwargs' for
                every data_var in to_plot.

        """
        if not to_plot:
            return {}

        data_vars_plot_kwargs = {}

        for k, v in to_plot.items():
            plot_kwargs = v.get("plot_kwargs", {})
            if not plot_kwargs.get("label"):
                plot_kwargs["label"] = v.get("label", k)
            data_vars_plot_kwargs[k] = plot_kwargs

        return data_vars_plot_kwargs

    def create_legend_handles(
        *, visualization_kwargs: dict, data_vars_plot_kwargs: dict
    ):
        """Creates legend handles

        Processes entries in data_vars_plot_kwargs (from to_plot) and
        visualization_kwargs.

        Args:
            visualization_kwargs (dict): The visualization kwargs
            data_vars_plot_kwargs (dict): The resolved entries in to_plot

        Returns:
            Tuple[list, list]: Tuple of legend handles and legend labels lists
                as required by ax.legend(handles, labels)
            data_vars_plot_kwargs: The updated entries data_vars_plot_kwargs
                as required by plot_attractor.
        """
        # Some defaults
        circle_kwargs = dict(xy=(0.5, 0.5), radius=0.25, edgecolor="none")
        rect_kwargs = dict(
            xy=(0.0, 0.0), height=0.75, width=1.0, edgecolor="none"
        )

        # Lists to be populated for matplotlib legend
        legend_handles = []
        legend_labels = []

        for k, kwargs in data_vars_plot_kwargs.items():
            label = kwargs.pop("label", kwargs)
            kwargs["linewidth"] = kwargs.get("linewidth", 0.0)
            data_vars_plot_kwargs[k] = kwargs

            # Determine color
            if "color" in kwargs:
                color = kwargs["color"]

            elif "cmap" in kwargs:
                cmap = mpl.cm.get_cmap(kwargs["cmap"])
                color = cmap(1.0)

            else:
                log.warning(f"No color defined for data_var '{k}'!")
                color = None

            # Create and add the handle and the label
            legend_handles.append(Circle(**circle_kwargs, facecolor=color))
            legend_labels.append(label)

        for k, kwargs in visualization_kwargs.items():
            if "to_plot" in kwargs:
                for dvar_name, dvar_kwargs in kwargs["to_plot"].items():
                    # Make sure a linewidth is set
                    dvar_kwargs["linewidth"] = dvar_kwargs.get(
                        "linewidth", 0.0
                    )
                    kwargs["to_plot"][dvar_name] = dvar_kwargs

                    # Create and add the handle and the label
                    legend_handles.append(
                        Rectangle(**rect_kwargs, **dvar_kwargs)
                    )
                    legend_labels.append(dvar_kwargs.get("label", dvar_name))
            else:
                kwargs["linewidth"] = kwargs.get("linewidth", 0.0)
                data_vars_plot_kwargs[k] = kwargs

                legend_handles.append(Rectangle(**rect_kwargs, **kwargs))
                legend_labels.append(kwargs.get("label", k))

        return [legend_handles, legend_labels], data_vars_plot_kwargs

    def apply_analysis_steps(
        data: xr.Dataset,
        analysis_steps: Sequence[Union[str, Tuple[str, str]]],
        *,
        analysis_funcs: dict,
        analysis_kwargs: dict,
    ):
        """Perform the sequence of analysis steps until the first conclusive.

        Args:
            data (xarray.Dataset): The data to analyse.
            analysis_steps (Sequence[Union[str, Tuple[str, str]]]): The analysis steps that are to be made until one is conclusive.
                Applied per universe.
            analysis_funcs (dict): The entries need to match the
                analysis_steps. Map of the analysis_steps to their Callables
            analysis_kwargs (dict): The entries need to match the
                analysis_steps. The subentry (dict) is passed on to the
                corresponding analysis function.

        Returns:
            analysis_key (str): The key of the conclusive analysis step.
            attractor (xarray.Dataset): The data corresponding to this analysis.
        """
        for analysis_key, analysis_func in analysis_steps:
            analysis_func_kwargs = analysis_kwargs.get(analysis_func, {})

            # resolve the analysis function from its name
            if isinstance(analysis_func, str):
                if analysis_func in analysis_funcs:
                    analysis_func = analysis_funcs[analysis_func]

                else:
                    # Try to get it from dataprocessing ... might fail.
                    analysis_func = getattr(
                        utopya.dataprocessing, analysis_func
                    )

            # Perfom the analysis step
            conclusive, attractor = analysis_func(data, **analysis_func_kwargs)

            # Return if conclusive
            if conclusive:
                return analysis_key, attractor

        # Return non-conclusive
        return None, None

    def resolve_rectangle(coord: dict, rectangles: xr.Dataset) -> Rectangle:
        """Resolve the rectangle patch at this coordinate

        Args:
            coord (dict): The bifurcation parameter's coordinate
            rectangles (xarray.Dataset): The rectangles that cover the 2D space
                spanned by the coordiantes. The `coord` should be one entry of
                rectangles.coords.

        Raises:
            ValueError: Coordinate not available in rectangles.

        Returns:
            Rectangle: A rectangle around a universe with coord and
                shape defined by rectangles.
        """
        try:
            rectangle = rectangles.sel(coord)

        except Exception as exc:
            raise ValueError(
                "The requested paramspace coordinate(s) {} are "
                "not coordinates of rectangles {}. Plot failed."
                "".format(coord, rectangles.coords)
            ) from exc

        rect_spec = rectangle["rect_spec"]
        return Rectangle(*rect_spec.item())

    def append_vis_patch(
        attrator_key: str,
        attractor: xr.Dataset,
        vis_patches: dict,
        vis_kwargs: dict,
        **resolve_rectangle_args,
    ):
        """Append visualization patch

        Performs postprocess for

            - attractor key 'fixpoint' and 'endpoint' if: 'to_plot' in entry of
                vis_kwargs. Then finds the data_var with highest valued
                datapoint.

        Args:
            attractor_key (str): Key according to which to decode the attractor
            attractor (xarray.Dataset): The Dataset with the encoded attractor
                information. See possible encodings
            vis_patches (dict): The map of attractor_key to
                List[Rectangle] where to append the new patch
            vis_kwargs (dict): The visualization kwargs
            **resolve_rectangle_args: Passed on to resolve_rectangle

        Raises:
            ValueError: Bad postprocess key

        Returns:
            vis_patches (dict): The new map of attractor_key to
                List[Rectangle]

        Deleted Parameters:
            resolve_rectangle_args (dict): Args as required by
                resolve_rectangle
        """
        # Depending on the kind of attractor, add different patches
        kwargs = vis_kwargs.get(attractor_key)
        if kwargs is None:
            return vis_patches

        # Postprocess fixpoint and to_plot, append rectangle
        if (
            attractor_key == "fixpoint" or attractor_key == "endpoint"
        ) and "to_plot" in kwargs:
            max_value = -np.inf
            for data_var_name, data_var in attractor.data_vars.items():
                if data_var.max() > max_value:
                    max_value = data_var.max()
                    max_name = data_var_name

            rect = resolve_rectangle(**resolve_rectangle_args)

            attractor_var_key = attractor_key + "_" + max_name
            vis_patches[attractor_var_key].append(rect)

        # Append rectangle
        elif vis_patches.get(attractor_key) is not None:
            rect = resolve_rectangle(coord=coord, rectangles=rects)
            vis_patches[attractor_key].append(rect)

        return vis_patches

    def append_plot_attractor(
        attractor_key: str,
        attractor: xr.Dataset,
        *,
        coord: float = None,
        scatter_kwargs: list,
        **plot_kwargs,
    ):
        """Resolves how to plot attractor of specified type
        at specific bifurcation parameter value.

        Args:
            attractor_key (str): Key according to which to decode the attractor
            attractor (xarray.Dataset): The Dataset with the encoded attractor
                information. See possible encodings
            coord (float, optional): The bifurcation parameter's coordinate,
                if None its derived from the attractors coordinates
            scatter_kwargs (list): The list of scatter datasets where to append
                the new scatter.
            plot_kwargs (dict, optional): The kwargs used to specify ax.scatter
                where the entries match the attractor.data_vars

        Possible encodings, i.e. values for ``attractor_key`` (of type
        :py:class:`xarray.Dataset` or a compatible type):

            - ``fixpoint``: dataset with dimensions ``()``
            - ``scatter``: dataset with dimensions ``(time: >=1)``
            - ``multistability``: dataset with dimensions
                ``(<initial_state>: >= 1)``
            - ``oscillation``: dataset with dimensions ``(osc: 2)``, the
                minimum and maximum

        .. note::

            The attractor must contain the bifurcation parameter coordinate

        Raises:
            KeyError: Unknown ``attractor_key`` or no bifurcation coordinate
            ValueError: Attractor encoding mismatched with the given
                ``attractor_key``

        Returns:
            dict: The new list of scatter datasets
        """
        # Get the bifurcation parameter coordinate
        if not coord:
            try:
                coord = attractor[dim]

            except KeyError as err:
                raise ValueError(
                    f"No bifurcation parameter coordinate '{dim}' "
                    "could be found! Either have it as a "
                    "coordinate in 'attractor' or pass it to "
                    "'plot_attractor' explicitly."
                ) from err

        # Resolve the scatter kwargs depending on attractor key
        if attractor_key in ("fixpoint", "endpoint", "multistability"):
            for data_var_name, data_var in attractor.data_vars.items():
                data_var = data_var.where(data_var != np.nan, drop=True)
                entries = 1
                if data_var.shape:
                    entries = len(data_var)

                scatter_kwargs.append(
                    dict(
                        x=[coord] * entries,
                        y=data_var,
                        **plot_kwargs.get(data_var_name, {}),
                    )
                )

        elif attractor_key == "scatter":
            for data_var_name, data_var in attractor.data_vars.items():
                if "cmap" in plot_kwargs.get(data_var_name, {}):
                    scatter_kwargs.append(
                        dict(
                            x=[coord] * len(data_var.data),
                            y=data_var,
                            c=attractor["time"],
                            **plot_kwargs.get(data_var_name, {}),
                        )
                    )

                else:
                    scatter_kwargs.append(
                        dict(
                            x=[coord] * len(data_var.data),
                            y=data_var,
                            **plot_kwargs.get(data_var_name, {}),
                        )
                    )

        elif attractor_key == "oscillation":
            for data_var_name, data_var in attractor.data_vars.items():
                scatter_kwargs.append(
                    dict(
                        x=[coord] * len(data_var.data),
                        y=data_var,
                        **plot_kwargs.get(data_var_name, {}),
                    )
                )

        elif attractor_key:
            raise ValueError(
                f"Invalid attractor-key '{attrator_key}'! "
                "Available keys: 'endpoint', fixpoint',"
                " 'multistability', 'scatter', 'oscillation'."
            )

        return scatter_kwargs

    def resolve_scatter(
        data: xr.Dataset, *, spin_up_time: int = 0, **kwargs
    ) -> tuple:
        """A mock analysis function to plot all times larger than a spin
        up time.

        Args:
            data (xarray.Dataset): The dataset to analyse
            spin_up_time (int, optional): The spin-up-time
            **kwargs: Ignored
        """
        return True, data.where(data.time >= spin_up_time, drop=True)

    # .........................................................................

    # Check argument values
    if not dim and not dims:
        raise ValueError(
            "No dim (str) or dims (Tuple[str, str]) specified. "
            "Use dim for a 1d-bifurcation diagram and dims for a "
            "2d-bifurcation diagram."
        )

    if dim and dims:
        raise ValueError(
            "dim='{}' and dims='{}' specified. "
            "Use either dim for a 1d-bifurcation diagram or dims "
            "for a 2d-bifurcation diagram."
            "".format(dim, dims)
        )

    if dims is not None and len(dims) != 2:
        raise ValueError(
            f"Argument dims should be of length 2, but was: {dims}"
        )

    # TODO In the future, consider not using `dim` below here but handling it
    #      via the length of `dims`.

    # Default values
    if visualization_kwargs is None:
        visualization_kwargs = {}

    if analysis_kwargs is None:
        analysis_kwargs = {}

    # Resolve legend handles and visualization kwargs
    data_vars_plot_kwargs = resolve_to_plot_kwargs(to_plot)
    legend_handles, data_vars_plot_kwargs = create_legend_handles(
        visualization_kwargs=visualization_kwargs,
        data_vars_plot_kwargs=data_vars_plot_kwargs,
    )

    # Define default analysis functions
    analysis_funcs = dict(
        endpoint=utdp.find_endpoint,
        fixpoint=utdp.find_fixpoint,
        multistability=utdp.find_multistability,
        oscillation=utdp.find_oscillation,
        scatter=resolve_scatter,
    )

    # If given, update
    if custom_analysis_funcs:
        log.debug("Updating with custom analysis functions ...")
        analysis_funcs = recursive_update(
            analysis_funcs, custom_analysis_funcs
        )

    analysis_steps = resolve_analysis_steps(analysis_steps)

    # Obtain the rectangles covering space spanned by the coordinates
    rectangle_map_kwargs = kwargs.get("rectangle_map_kwargs", {})
    if dim:
        rects, limits = calc_pxmap_rectangles(
            x_coords=mv_data[dim].values, y_coords=None, **rectangle_map_kwargs
        )
    elif dims:
        rects, limits = calc_pxmap_rectangles(
            x_coords=mv_data[dims[0]].values,
            y_coords=mv_data[dims[1]].values,
            **rectangle_map_kwargs,
        )

    # Obtain the list of param_coords to iterate
    if dim:
        param_iter = mv_data[dim].values

    elif dims:
        param_iter = itertools.product(
            mv_data[dims[0]].values, mv_data[dims[1]].values
        )

    # Map of analysis_key to list[mpatch.Rectangle]
    vis_patches = {}
    for analysis_key, _ in analysis_steps:
        if not visualization_kwargs.get(analysis_key):
            continue

        if "to_plot" in visualization_kwargs[analysis_key]:
            for var_key, _ in visualization_kwargs[analysis_key][
                "to_plot"
            ].items():
                analysis_var_key = analysis_key + "_" + var_key
                vis_patches[analysis_var_key] = []

        else:
            vis_patches[analysis_key] = []

    # The List[dict] passed to ax.scatter
    scatter_kwargs = []
    scatter_coords_kwargs = []

    # Iterate the parameter coordinates
    for param_coord in param_iter:
        # Resolve the param_coord to dict
        if dim:
            param_coord = {dim: param_coord}

        elif dims:
            param_coord = {dims[0]: param_coord[0], dims[1]: param_coord[1]}

        # Plot coord
        if dim and kwargs.get("plot_coords_kwargs"):
            plot_coords_kwargs = kwargs.get("plot_coords_kwargs")
            scatter_coords_kwargs.append(
                {
                    "x": param_coord[dim],
                    "y": plot_coords_kwargs.pop("y", 0.0),
                    **plot_coords_kwargs,
                }
            )

        if dims and kwargs.get("plot_coords_kwargs"):
            scatter_coords_kwargs.append(
                {
                    "x": param_coord[dims[0]],
                    "y": param_coord[dims[1]],
                    **kwargs.get("plot_coords_kwargs"),
                }
            )

        # Select the data and analyse
        data = mv_data.sel(param_coord)
        attractor_key, attractor = apply_analysis_steps(
            data,
            analysis_steps,
            analysis_funcs=analysis_funcs,
            analysis_kwargs=analysis_kwargs,
        )

        # If conclusive, append a rectangular patch to the attractor_key's
        # patch collection
        if attractor_key:
            # Determine coordinate value
            if dim:
                rect_map_kwargs = kwargs.get("rectangle_map_kwargs", {})
                y = rect_map_kwargs.get("default_pos", (0.0, 0.0))[1]
                coord = dict(x=param_coord[dim], y=y)

            elif dims:
                coord = dict(x=param_coord[dims[0]], y=param_coord[dims[1]])

            vis_patches = append_vis_patch(
                attractor_key,
                attractor,
                vis_patches,
                visualization_kwargs,
                coord=coord,
                rectangles=rects,
            )

        # For 1d case ...
        if dim and to_plot:
            scatter_kwargs = append_plot_attractor(
                attractor_key,
                attractor,
                coord=param_coord[dim],
                scatter_kwargs=scatter_kwargs,
                **data_vars_plot_kwargs,
            )

    # Draw collection of visualization patches
    for analysis_key, _ in analysis_steps:
        if not visualization_kwargs.get(analysis_key):
            continue

        if analysis_key in visualization_kwargs:
            vis_kwargs = visualization_kwargs[analysis_key]

            if "to_plot" in vis_kwargs:
                for var_key, var_kwargs in vis_kwargs["to_plot"].items():
                    attractor_var_key = analysis_key + "_" + var_key

                    pc = PatchCollection(
                        vis_patches[attractor_var_key], **var_kwargs
                    )
                    hlpr.ax.add_collection(pc)

            else:
                pc = PatchCollection(vis_patches[analysis_key], **vis_kwargs)
                hlpr.ax.add_collection(pc)

        # else: nothing to do

    # Scatter the universe's coordinates
    for kws in scatter_coords_kwargs:
        hlpr.ax.scatter(**kws)

    # Scatter the attractor
    for kws in scatter_kwargs:
        hlpr.ax.scatter(**kws)

    # Provide PlotHelper defaults
    hlpr.provide_defaults("set_limits", **limits)

    if dim:
        hlpr.provide_defaults("set_labels", x=dim, y="state")
    elif dims:
        hlpr.provide_defaults("set_labels", x=dims[0], y=dims[1])

    if legend_handles:
        hlpr.ax.legend(
            legend_handles[0],
            legend_handles[1],
            handler_map={Circle: HandlerEllipse()},
            **kwargs.get("legend_kwargs", {}),
        )



"""Implements the `utopya test` subcommand"""

import os
import sys

import click

from ._shared import OPTIONS, add_options, complete_model_names
from ._utils import Echo

# -----------------------------------------------------------------------------


@click.command(
    "test",
    help=(
        "Run the Python model tests associated with the specified model.\n"
        "\n"
        "Looks up the test directory, temporarily moves to that directory, "
        "and invokes pytest there. Any additional pytest arguments are passed "
        "through. See pytest docs for more information."
    ),
    context_settings=dict(ignore_unknown_options=True),
)
#
# Select a model
@click.argument("model_name", shell_complete=complete_model_names)
@add_options(OPTIONS["label"])
#
# Pass pytest arguments through
@click.argument("pytest_args", nargs=-1)
#
# NOTE This function needs a name that is not simply ``test`` because pytest
#      will also search this file for test discovery and we cannot have a
#      function here that looks like a test function.
def run_test(
    model_name: str,
    label: str,
    pytest_args: tuple,
):
    """Invokes the associated Python tests for a model using pytest."""
    import pytest

    import utopya

    _log = utopya._getLogger("utopya")

    # Get model information and the corresponding test directory
    model = utopya.Model(name=model_name, bundle_label=label)
    model_info = model.info_bundle

    py_tests_dir = model_info.paths.get("py_tests_dir")
    if not py_tests_dir:
        _log.caution("No tests defined for this model.")
        return

    py_tests_dir = os.path.abspath(py_tests_dir)

    # Add the model tests' parent directories to the PATH to allow imports
    prepend_to_sys_path = (
        os.path.dirname(os.path.dirname(py_tests_dir)),
        os.path.dirname(py_tests_dir),
    )
    for p in prepend_to_sys_path:
        sys.path.insert(0, p)

    # Move to the test directory ...
    old_wd = os.getcwd()
    os.chdir(py_tests_dir)

    # ... and invoke the tests
    _log.progress("Invoking associated Python model tests ...")
    _log.remark(
        "Temporarily setting working directory to model test directory:\n  %s",
        py_tests_dir,
    )
    _log.remark("Full test command:\n\n  pytest %s\n\n", " ".join(pytest_args))
    try:
        sys.exit(pytest.main(list(pytest_args)))

    finally:
        # Change back to previous working directory to not have side effects
        os.chdir(old_wd)

        # Remove previously prependend directories from system PATH
        for p in prepend_to_sys_path:
            sys.path.remove(p)



"""Tests shell completion features (indirectly)"""

import pytest

from utopya_cli._shared import *

from .. import ADVANCED_MODEL, DUMMY_MODEL, TEST_PROJECT_NAME
from .._fixtures import *
from . import invoke_cli
from .test_run_and_eval import _check_result


class MockContext:
    def __init__(self, **kwargs):
        for key, val in kwargs.items():
            setattr(self, key, val)


# -----------------------------------------------------------------------------


def test_complete_model_names(with_test_models):
    """Tests auto-completion of model names"""
    ctx = None
    param = None
    complete = complete_model_names

    assert DUMMY_MODEL in complete(ctx, param, "")
    assert ADVANCED_MODEL in complete(ctx, param, "")

    # With a (long enough) incomplete string, will reduce the list
    incomplete = ADVANCED_MODEL[:5]
    assert complete(ctx, param, incomplete) == [ADVANCED_MODEL]


def test_complete_project_names(with_test_models):
    """Tests auto-completion of project names"""
    ctx = None
    param = None
    complete = complete_project_names

    assert TEST_PROJECT_NAME in complete(ctx, param, "")

    # With a (long enough) incomplete string, will reduce the list
    incomplete = TEST_PROJECT_NAME[:10]
    assert complete(ctx, param, incomplete) == [TEST_PROJECT_NAME]


def test_complete_run_dirs(with_test_models, tmp_output_dir):
    """Tests auto-completion of model names"""
    ctx = MockContext(params=dict(model_name="invalid model name"))
    param = None
    complete = complete_run_dirs

    # Should be empty for invalid model names
    assert not complete(ctx, param, "")

    # Now use a valid model name to generate some output
    # Need to make some runs first to have some output
    res = invoke_cli(("run", DUMMY_MODEL, "-d"))
    _check_result(res, expected_exit=0)

    res = invoke_cli(
        ("run", DUMMY_MODEL, "--no-eval", "--note", "some_note", "-d")
    )
    _check_result(res, expected_exit=0)

    # Now check if there are completed directories available
    # NOTE Need to set `extra_search_dirs` because otherwise would create side
    #      effects from the test or miss the temporary directory
    ctx = MockContext(params=dict(model_name=DUMMY_MODEL))
    c1 = complete(ctx, None, "")
    c2 = complete(ctx, None, "", extra_search_dirs=[str(tmp_output_dir)])
    assert len(c1) <= len(c2)

    # Check that the latest run directory is part of the suggestions.
    # Need to (tediously) parse the CLI output to get to the run directory.
    IGNORE = ("eval", ".h5", ".yml")
    run_dir = [
        line.strip()
        for line in res.output.split("\n")
        if "_some_note" in line and not any(s in line for s in IGNORE)
    ][0]
    # assert os.path.basename(run_dir) not in c1 # FIXME test has wrong out_dir
    assert os.path.basename(run_dir) in c2



"""Implements the `utopya config` subcommand tree of the CLI"""

import os
import sys
from typing import Tuple

import click

from ._utils import Echo, set_entries_from_kv_pairs


@click.command(
    help=(
        "Read and modify the utopya configuration.\n"
        "\n"
        "Configuration entries are grouped into multiple categories, which "
        "each category corresponding to a YAML file within the utopya "
        "config directory."
        "The ``utopya config`` command allows revealing the configuration "
        "file, opening it in an editor, or setting values directly via the "
        "command line interface.\n"
        "\n"
        "Entries within these configuration files can be set using the "
        "``--set`` and ``KV_PAIRS`` argument, expecting ``key=value`` pairs. "
        "The key may be a dot-separated string of keys for dict traversal, "
        "like ``foo.bar.baz``. "
        "To delete entries, set the value to ``DEL``. For example:\n"
        "\n"
        "    ``utopya config user --get --set foo.bar=spam bar.baz=DEL``\n"
        "\n"
        "This will set the user configuration entry ``foo.bar``, delete the "
        "entry ``bar.baz``, and then retrieve the new configuration. "
        "To include spaces, put the whole key-value pair into quotes "
        "(``'key=value'``); everything after the ``=`` is parsed as a string "
        "anyway."
        "If the configuration file does not exist, it will be created."
    )
)
@click.argument(
    "cfg_name",
    type=click.Choice(
        (
            "user",
            "utopya",
            "batch",
        )  # TODO load these from utopya if import time allows
    ),
)
@click.argument(
    "kv_pairs",
    default=None,
    nargs=-1,
    type=str,
)
@click.option(
    "-R",
    "--reveal",
    is_flag=True,
    help=(
        "Reveals the location of the configuration files, then exits. "
        "If the config file does not yet exist, deploys an empty file at "
        "the expected location."
    ),
)
@click.option(
    "--set",
    "set_entry",
    is_flag=True,
    help=(
        "Set entries in the specified configuration, using the KV_PAIRS. "
        "This option is mutually exclusive with the ``--edit`` option."
    ),
)
@click.option(
    "--edit",
    is_flag=True,
    help=(
        "Open the selected config file in an editor. "
        "This option is mutually exclusive with the ``--set`` option."
    ),
)
@click.option(
    "--get",
    "get_entry",
    is_flag=True,
    help=(
        "Retrieve all entries from the specified configuration. This is "
        "always invoked *after* the ``--set`` command was executed (if given)."
    ),
)
@click.option(
    "--accept-yaml/--no-yaml",
    "allow_yaml",
    is_flag=True,
    default=True,
    help=(
        "Whether to accept YAML syntax for parsing the values in ``KV_PAIRS`` "
        "(default: ``true``). "
        "This allows specifying list- or dict-like values. In case this "
        "results in a dict, the object at the specified key will be "
        "recursively updated"
    ),
)
def config(
    *,
    cfg_name: str,
    kv_pairs: Tuple[str],
    reveal: bool,
    set_entry: bool,
    edit: bool,
    get_entry: bool,
    allow_yaml: bool,
):
    """Sets and reads configuration entries"""

    if not reveal and not get_entry and not (set_entry or edit):
        Echo.error(
            "Need at least one of the options "
            "--get, --set, --edit, or --reveal!"
        )
        Echo.help(exit=1)

    elif set_entry and edit:
        Echo.error("Can only pass one of --set or --edit, not both!")
        Echo.help(exit=1)

    elif set_entry and not kv_pairs:
        Echo.error("Got --set option but the KV_PAIRS argument is missing!")
        Echo.help(exit=1)

    from utopya.cfg import get_cfg_path, load_from_cfg_dir, write_to_cfg_dir

    # -- Reveal
    # NOTE The launch command is only invoked if not in a test context
    if reveal:
        cfg_fpath = get_cfg_path(cfg_name)
        Echo.info(f"Revealing '{cfg_name}' config file ...")
        Echo.remark(f"  Location:  {cfg_fpath}")

        if not os.path.exists(cfg_fpath):
            write_to_cfg_dir(cfg_name, {})

        rv = 0
        if not os.environ.get("PYTEST_CURRENT_TEST"):
            rv = click.launch(cfg_fpath, locate=True)
        sys.exit(rv)

    # -- Writing/editing entries
    if set_entry:
        Echo.note(
            f"Setting '{cfg_name}' config values from key-value pairs ..."
        )
        cfg = load_from_cfg_dir(cfg_name)  # empty dict if file is missing
        set_entries_from_kv_pairs(
            *kv_pairs, add_to=cfg, _log=Echo, allow_yaml=allow_yaml
        )
        write_to_cfg_dir(cfg_name, cfg)
        Echo.progress(
            "Set %d entr%s in '%s' configuration.\n",
            len(kv_pairs),
            "ies" if len(kv_pairs) != 1 else "y",
            cfg_name,
        )

    elif edit:
        Echo.note(f"Opening editor with '{cfg_name}' config file ...")
        try:
            click.edit(filename=get_cfg_path(cfg_name), extension=".yml")

        except Exception as exc:
            Echo.error(f"Editing config file '{cfg_name}' failed!", error=exc)
            sys.exit(1)

        else:
            Echo.success(f"Config file '{cfg_name}' edited successfully.")

    # -- Reading entries
    if get_entry:
        from utopya.tools import pformat

        Echo.note(
            "Reading '%s' config (%s) ...", cfg_name, get_cfg_path(cfg_name)
        )
        cfg = load_from_cfg_dir(cfg_name)
        Echo.info(
            "\n--- {} Configuration ---\n\n{}".format(
                cfg_name.replace("_", " ").title(), pformat(cfg)
            )
        )



"""This module implements various generic tools"""

import importlib
import os
import sys
from types import ModuleType
from typing import Any, Union

from .logging import backend_logger as _log

# -- YAML ---------------------------------------------------------------------


def load_cfg_file(fpath: str, *, loader: str = None) -> Any:
    """Loads a configuration file from the given file. Allows to automatically
    determine which kind of loading function to use.

    Currently supported loading functions: YAML

    Args:
        fpath (str): Path to the configuration file to load
        loader (str, optional): Name of the loader to use. If not given, will
            determine it from the file extension of ``fpath``.

    Returns:
        Any: The return value of the load function.

    Raises:
        ValueError: On invalid ``loader`` argument or a file extension that
            does not map to a supported loader.
    """
    if loader is None:
        # Determine from file extension
        _, ext = os.path.splitext(fpath)
        loader = ext.replace(".", "")
    loader = loader.lower()

    # Determine load function
    # TODO add toml, json, …
    if loader in ("yaml", "yml"):
        from paramspace.yaml import yaml

        load = yaml.load

    else:
        raise ValueError(
            f"Unsupported loader '{loader}' for configuration file! Check the "
            "file path and the `loader` argument.\n"
            f"  File path:  {fpath}\n"
            f"  Available loaders:  yml, yaml"
        )

    # Load now
    with open(os.path.expanduser(fpath), "r") as f:
        return load(f)


# -----------------------------------------------------------------------------


def import_package_from_dir(
    mod_dir: str, *, mod_str: str = None
) -> ModuleType:
    """Helper function to import a package-like module that is importable only
    when adding the module's parent directory to :py:data:`sys.path`.

    The ``mod_dir`` directory needs to contain an ``__init__.py`` file.
    If that is not the case, you cannot use this function, because the
    directory does not represent a package.

    .. hint::

        This function is very useful to get access to a local package that is
        *not* installed, as might be the case for your model implementation.
        Assuming you have an ``impl`` package right beside the current
        ``__file__`` and that package includes your ``Model`` class
        implementation:

        .. code-block:: text

            - run_model.py         # Current __file__
            - impl/                # Implementation package
              |-- __init__.py      # Exposes impl.model.Model
              |-- model.py         # Implements Model class
              |-- ...

        You can get access to it like this from within ``run_model.py``:

        .. code-block:: python

            import os
            from utopya_backend import import_package_from_dir

            impl = import_package_from_dir(
                os.path.join(os.path.dirname(__file__), "impl")
            )
            Model = impl.Model

    Args:
        mod_dir (str): Path to the module's root *directory*, ``~`` expanded.
            For robustness, relative paths are *not* allowed.
        mod_str (str, optional): Name under which the module can be imported
            with the *parent* of  ``mod_dir`` being in :py:data:`sys.path`.
            If not given, will assume it is equal to the last segment of
            ``mod_dir``.

    Returns:
        ModuleType: The imported module.

    Raises:
        ImportError: If ``debug`` is set and import failed for whatever reason
        FileNotFoundError: If ``mod_dir`` did not point to an existing
            *directory*
    """
    mod_dir = os.path.expanduser(mod_dir)
    if not os.path.isabs(mod_dir):
        raise ValueError(
            f"Need an absolute path for argument `mod_dir` but got:  {mod_dir}"
        )

    elif not os.path.isdir(mod_dir):
        raise FileNotFoundError(
            "The `mod_dir` argument to import a module from a path should be "
            f"the path to an existing directory! Given path:  {mod_dir}"
        )

    # Normalize it to ensure that it does not have a trailing slash
    mod_dir = os.path.realpath(mod_dir)

    # May need to infer module string
    if mod_str is None:
        mod_str = os.path.basename(mod_dir)

    # Need the parent directory in the path, because the import is only
    # possible from there. This, in turn, depends on the depth of the module
    # string, so the parent directory should be chosen accordingly.
    mod_parent_dir = mod_dir
    for _ in mod_str.split("."):
        mod_parent_dir = os.path.dirname(mod_parent_dir)

    _log.debug("Importing module '%s' from directory ...", mod_str)
    try:
        sys.path.insert(0, mod_parent_dir)
        mod = importlib.import_module(mod_str)

    except Exception as exc:
        raise ImportError(
            f"Failed importing module '{mod_str}'!\n"
            f"Make sure that {mod_dir}/__init__.py can be loaded without "
            "errors (with its parent directory being part of sys.path) "
            "and that the `mod_str` argument is correct; if you did not "
            "specify `mod_str` explicitly, consider doing so."
        ) from exc

    _log.debug("Successfully imported module from directory:  %s", mod)
    return mod



"""Implementation of the utopya project registry"""

import logging
import os
from typing import Any, Dict, List, Optional

from pydantic import DirectoryPath, FilePath

from ._yaml_registry import BaseSchema, RegistryEntry, YAMLRegistry
from .cfg import PROJECT_INFO_FILE_SEARCH_PATHS, UTOPYA_CFG_SUBDIRS
from .tools import load_yml, recursive_update

log = logging.getLogger(__name__)

# -- Schema definition --------------------------------------------------------
# .. Sub-schemas ..............................................................


class ProjectPaths(BaseSchema):
    """Schema to use for a project's ``paths`` field"""

    base_dir: DirectoryPath
    project_info: Optional[FilePath] = None
    models_dir: Optional[DirectoryPath] = None
    py_tests_dir: Optional[DirectoryPath] = None
    py_plots_dir: Optional[DirectoryPath] = None
    mv_project_cfg: Optional[FilePath] = None
    project_base_plots: Optional[FilePath] = None

    # TODO What about the BatchTaskManager configuration?


class ProjectMetadata(BaseSchema):
    """Schema to use for a project's ``metadata`` field"""

    version: Optional[str] = None
    long_name: Optional[str] = None
    description: Optional[str] = None
    long_description: Optional[str] = None
    license: Optional[str] = None
    authors: Optional[List[str]] = None
    email: Optional[str] = None
    website: Optional[str] = None
    utopya_compatibility: Optional[str] = None
    language: Optional[str] = None
    requirements: Optional[List[str]] = None
    misc: Optional[Dict[str, Any]] = None


class ProjectSettings(BaseSchema):
    """Schema to use for a project's ``settings`` field"""

    preload_project_py_plots: Optional[bool] = None
    """Whether to preload the project-level plot module (``py_plots_dir``)
    after initialization of the :py:mod:`~utopya.eval.plotmanager.PlotManager`.
    If not given, will load the module.
    """

    preload_framework_py_plots: Optional[bool] = None
    """Whether to preload the framework-level plot module (``py_plots_dir``)
    after initialization of the :py:mod:`~utopya.eval.plotmanager.PlotManager`.
    If not given, will load the module.
    """


# .............................................................................


class ProjectSchema(BaseSchema):
    """The data model for a project registry entry"""

    project_name: str
    framework_name: Optional[str] = None
    paths: ProjectPaths
    metadata: ProjectMetadata
    settings: ProjectSettings = {}
    run_cfg_format: str = "yaml"
    cfg_set_abs_search_dirs: Optional[List[str]] = None
    cfg_set_model_source_subdirs: Optional[List[str]] = None
    custom_py_modules: Optional[Dict[str, DirectoryPath]] = None
    output_files: Optional[dict] = None  # TODO Needs sub-schema
    debug_level_updates: Optional[Dict[str, dict]] = None  # TODO Implement


# -- Project ------------------------------------------------------------------


class Project(RegistryEntry):
    """A registry entry that describes a project"""

    SCHEMA: type = ProjectSchema

    @property
    def framework_project(self) -> Optional["Project"]:
        """If a framework project is defined, retrieve it from the registry"""
        if not self.framework_name:
            return None

        from . import PROJECTS

        return PROJECTS[self.framework_name]

    def get_git_info(self, *, include_patch_info: bool = False) -> dict:
        """Returns information about the state of this project's git repository
        using the ``python-git-info`` package.

        If no git information is retrievable, e.g. because the project's
        ``base_dir`` does not contain a git repository, will still return a
        dict but with ``have_git_info`` entry set to False.

        Otherwise the git information will be in the ``latest_commit`` entry.

        Args:
            include_patch_info (bool, optional): If True, will attempt a
                subprocess call to ``git`` and store patch information
                alongside in the ``diff`` entry. In that case, the ``dirty``
                entry will denote whether there were uncommitted changes.

        Returns:
            dict: A dict containing information about the associated git repo.
        """
        import subprocess

        import gitinfo  # PyPI package `python-git-info`

        base_dir = str(self.paths.base_dir)
        sp_kws = dict(cwd=base_dir, capture_output=True, text=True)

        d = dict(
            project_name=self.project_name,
            project_base_dir=base_dir,
            have_git_repo=False,
            latest_commit=None,
            dirty="unknown",
            git_status=[],
            git_diff="",
        )

        # Get git information (without requiring git)
        info = gitinfo.get_git_info(base_dir)
        if info:
            d["have_git_repo"] = True
            d["latest_commit"] = info

        if info and include_patch_info:
            # Attempt subprocess git calls to find out more.
            # Make sure these commands work in the first place.
            try:
                git_status = subprocess.run(
                    ["git", "status", "--short"], **sp_kws, check=True
                )

            except Exception as exc:
                log.caution(
                    "Failed retrieving git patch information for "
                    f"{self.project_name} because git invocation via a python "
                    f"subprocess failed:\n\n{type(exc).__name__}: {exc}"
                )

            else:
                # git command works, can store information
                # git status
                git_status_stdout = git_status.stdout.strip()
                if git_status_stdout:
                    d["git_status"] = [
                        f.split(" ", 1) for f in git_status_stdout.split("\n")
                    ]

                # git diff
                git_diff_p = subprocess.run(["git", "diff", "-p"], **sp_kws)
                d["git_diff"] = git_diff_p.stdout.strip()

                # ... whether the repo has uncommitted changes
                git_diff = subprocess.run(["git", "diff", "--quiet"], **sp_kws)
                d["dirty"] = git_diff.returncode != 0

        return d


# -- ProjectRegistry ----------------------------------------------------------


class ProjectRegistry(YAMLRegistry):
    """The project registry"""

    def __init__(self, registry_dir: str = None):
        """Initializes the project registry, loading available entries from
        the registry directory in the utopya config directory.

        This also creates the ``projects`` directory, if not created yet.

        Args:
            registry_dir (str, optional): A custom projects
        """

        if registry_dir is None:
            registry_dir = UTOPYA_CFG_SUBDIRS["projects"]

        if not os.path.exists(registry_dir):
            os.makedirs(registry_dir)

        super().__init__(Project, registry_dir=registry_dir)

    def register(
        self,
        *,
        base_dir: str,
        info_file: str = None,
        custom_project_name: str = None,
        require_matching_names: bool = None,
        exists_action: str = "raise",
    ) -> Project:
        """Register or update information of a project.

        Args:
            base_dir (str): Project base directory
            info_file (str, optional): Path to info file which contains further
                path information and metadata (may be relative to base
                directory). If not given, will use some defaults to search for
                it.
            custom_project_name (str, optional): Custom project name,
                overwrites the one given in the info file
            require_matching_names (bool, optional): If set, will require that
                the custom project name is equal to the one given in the
                project info file. This allows checking that the file content
                does not diverge from some outside state.
            exists_action (str, optional): Action to take upon existing project

        Returns:
            Project: Project information for the new or validated project
        """
        log.info("Commencing project registration ...")

        # Parse base directory and info file; if not given, search for it
        base_dir = os.path.abspath(os.path.expanduser(base_dir))
        log.remark("Project base directory:\n  %s", base_dir)

        if info_file:
            info_file = os.path.realpath(
                os.path.join(base_dir, os.path.expanduser(info_file))
            )
            log.remark("Have project info file available:\n  %s", info_file)

        else:
            for _relpath in PROJECT_INFO_FILE_SEARCH_PATHS:
                _abspath = os.path.join(base_dir, _relpath)
                if os.path.isfile(_abspath):
                    info_file = _abspath
                    log.remark("Found project info file:\n  %s", info_file)
                    break

            else:
                _locs = "\n".join(
                    f"  - {p}" for p in PROJECT_INFO_FILE_SEARCH_PATHS
                )
                raise ValueError(
                    "Missing project info file!\nEither explicitly supply a "
                    "path or add one at the following search locations "
                    f"relative to the base directory:\n{_locs}"
                )

        # Load and parse the info file
        log.note("Loading project information from file ...")

        project_info = dict(paths=dict(base_dir=base_dir))
        _project_info = load_yml(info_file)
        _project_info = _project_info if _project_info else {}

        # Warn if base_dir was given in info file; the base directory given
        # explicitly should always take precedence
        if _project_info.get("paths", {}).get("base_dir"):
            log.caution(
                "The project info file contains a `paths.base_dir` entry, "
                "which will be ignored!"
            )

        project_info = recursive_update(_project_info, project_info)
        project_info["paths"]["project_info"] = info_file

        # Regularize paths, evaluating relative ones towards the base directory
        for path_name, path in project_info["paths"].items():
            if path_name in ("base_dir", "project_info"):
                continue

            project_info["paths"][path_name] = os.path.realpath(
                os.path.join(base_dir, path)
            )

        # May want to use a custom project name
        if custom_project_name:
            _project_name = project_info.get("project_name")
            if require_matching_names and custom_project_name != _project_name:
                raise ValueError(
                    "The custom project name '{}' does not match the name "
                    "given in the project info file, '{}'! "
                    "Either ensure that the names match exactly or unset the "
                    "`require_matching_names` flag.".format(
                        custom_project_name, project_info["project_name"]
                    )
                )

            project_info["project_name"] = custom_project_name
            log.remark("Using a custom project name:  %s", custom_project_name)
            log.remark(
                "Note that this creates a mismatch between the project "
                "info file and the registered project!"
            )

        # Now, create a registry entry from it
        log.note(
            "Preparing to add entry (exists_action: %s) ...", exists_action
        )
        project_name = project_info["project_name"]

        project = self.add_entry(
            project_name, exists_action=exists_action, **project_info
        )

        log.success(
            "Successfully stored or updated '%s' project information.",
            project_name,
        )
        log.remark(
            "NOTE: Changes to the project info file are *not* automatically "
            "tracked!\nTo update project information, repeat this procedure."
        )

        return self[project_name]


# -----------------------------------------------------------------------------

PROJECTS = ProjectRegistry()
"""The package-wide project registry"""



"""Provides the :py:class:`.Model` to work interactively with registered
utopya models"""

import glob
import logging
import os
import warnings
from tempfile import TemporaryDirectory
from typing import Dict, List, Tuple

from dantro.tools import adjusted_log_levels as _adjusted_log_levels
from dantro.tools import make_columns as _make_columns

from .cfg import load_from_cfg_dir
from .eval import DataManager
from .model_registry import ModelInfoBundle, get_info_bundle, load_model_cfg
from .multiverse import FrozenMultiverse, Multiverse

log = logging.getLogger(__name__)

# -----------------------------------------------------------------------------


class Model:
    """A class to work with Utopia models interactively.

    It attaches to a certain model and makes it easy to load config files,
    create a Multiverse from them, run it, and work with it further...
    """

    CONFIG_SET_MODEL_SOURCE_SUBDIRS = ("cfgs", "cfg_sets", "config_sets")
    """Directories within the model source directories to search through when
    looking for configuration sets. These are *not* used if the utopya config
    contains an entry overwriting this."""

    def __init__(
        self,
        *,
        name: str = None,
        info_bundle: ModelInfoBundle = None,
        bundle_label: str = None,
        base_dir: str = None,
        sim_errors: str = None,
        use_tmpdir: bool = False,
    ):
        """Initialize the ModelTest for the given model name

        Args:
            name (str, optional): Name of the model to attach to. If not
                given, need to pass info_bundle.
            info_bundle (ModelInfoBundle, optional): The required information
                to work with this model. If not given, will attempt to find the
                model in the model registry via ``name`` or ``bundle_label``.
            bundle_label (str, optional): A label to use for identifying the
                info bundle.
            base_dir (str, optional): For convenience, can specify this path
                which will be seen as the base path for config files; if set,
                arguments that allow specifying configuration files can specify
                them relative to this directory.
            sim_errors (str, optional): Whether to raise errors from Multiverse
            use_tmpdir (bool, optional): Whether to use a temporary directory
                to write data to. The default value can be set here; but the
                flag can be overwritten in the create_mv and create_run_load
                methods. For ``false``, the regular model output
                directory is used.

        Raises:
            ValueError: Upon bad ``base_dir``
        """
        # First, determine which model info bundle to use
        self._info_bundle = get_info_bundle(
            model_name=name, info_bundle=info_bundle, bundle_label=bundle_label
        )
        log.progress("Initializing Model instance for '%s' ...", self.name)

        # Show some project and framework information
        fw_name = None
        if self._info_bundle.project:
            fw_name = self._info_bundle.project.framework_name

        log.note("  Project:        %s", self._info_bundle.project_name)
        log.note("  Framework:      %s", fw_name)

        # Store other attributes
        self._sim_errors = sim_errors
        self._use_tmpdir = use_tmpdir

        self._base_dir = ""
        if base_dir:
            base_dir = os.path.expanduser(base_dir)

            if not os.path.isabs(base_dir):
                raise ValueError(
                    "Given base_dir path {} should be absolute, "
                    "but was not!".format(base_dir)
                )

            elif not os.path.exists(base_dir) or not os.path.isdir(base_dir):
                raise ValueError(
                    "Given base_dir path {} does not seem to "
                    "exist or is not a directory!"
                    "".format(base_dir)
                )

            self._base_dir = base_dir

        # Need to store Multiverses etc. such that they don't go out of scope
        self._mvs = []

    def __str__(self) -> str:
        """Returns an informative string for this Model instance"""
        return f"<Utopia '{self.name}' model>"

    # Properties ..............................................................

    @property
    def info_bundle(self) -> ModelInfoBundle:
        """The model info bundle"""
        return self._info_bundle

    @property
    def name(self) -> str:
        """The name of this Model object, which is at the same time the name
        of the attached model.
        """
        return self.info_bundle.model_name

    @property
    def base_dir(self) -> str:
        """Returns the path to the base directory, if set during init.

        This is the path to a directory from which config files can be loaded
        using relative paths.
        """
        return self._base_dir

    @property
    def default_model_cfg(self) -> dict:
        """Returns the default model configuration by loading it from the file
        specified in the info bundle.
        """
        cfg, _, _ = load_model_cfg(info_bundle=self.info_bundle)
        return cfg

    @property
    def default_config_set_search_dirs(self) -> List[str]:
        """Returns the default config set search directories for this model
        in the order of precedence:

        - defined on the project-level via ``cfg_set_abs_search_dirs``;
          these may also be format strings supporting the following set of
          keys: ``model_name``, ``project_base_dir``, and ``model_source_dir``
          (if set). If no project is associated, there will be no additional
          search directories.
        - names of subdirectories relative to the model source directory,
          defined in ``cfg_set_model_source_subdirs``. If no model source
          directory is known, no search directories will be added. If no
          project is associated, a standard set of search directories is
          used: ``cfgs``, ``cfg_sets``, ``config_sets``, as defined in
          :py:attr:`.CONFIG_SET_MODEL_SOURCE_SUBDIRS`.

        .. note::

            The output *may* contain relative paths.
        """
        search_dirs = []

        project = self.info_bundle.project
        model_source_dir = self.info_bundle.paths.get("source_dir")

        # Project-level search directories, potentially format strings
        if project and project.cfg_set_abs_search_dirs:
            format_kwargs = dict(
                project_base_dir=project.paths.base_dir,
                model_name=self.name,
            )
            if model_source_dir:
                format_kwargs["model_source_dir"] = model_source_dir

            search_dirs += [
                d.format(**format_kwargs)
                for d in project.cfg_set_abs_search_dirs
            ]

        # Relative to model source directory
        if model_source_dir:
            if project and project.cfg_set_model_source_subdirs:
                subdirs = project.cfg_set_model_source_subdirs
            else:
                subdirs = self.CONFIG_SET_MODEL_SOURCE_SUBDIRS

            for subdir in subdirs:
                search_dirs.append(os.path.join(model_source_dir, subdir))

        return search_dirs

    @property
    def default_config_sets(self) -> Dict[str, dict]:
        """Config sets at the default search locations.

        To retrieve an *individual* config set, consider using
        :py:meth:`~utopya.model.Model.get_config_set` instead of this property.

        For more information, see :ref:`config_sets`.
        """
        return self.get_config_sets(
            search_dirs=self.default_config_set_search_dirs
        )

    # Simulation control ......................................................

    def create_mv(
        self,
        *,
        from_cfg: str = None,
        from_cfg_set: str = None,
        run_cfg_path: str = None,
        use_tmpdir: bool = None,
        **update_meta_cfg,
    ) -> Multiverse:
        """Creates a :py:class:`utopya.multiverse.Multiverse` for this model,
        optionally loading a configuration from a file and updating it with
        further keys.

        Args:
            from_cfg (str, optional): The name of the config file (relative to
                the base directory) to be used.
            from_cfg_set (str, optional): Name of the config set to retrieve
                the run config from. Mutually exclusive with ``from_cfg`` and
                ``run_cfg_path``.
            run_cfg_path (str, optional): The path of the run config to use.
                Can not be passed if ``from_cfg`` or ``from_cfg_set`` arguments
                were given.
            use_tmpdir (bool, optional): Whether to use a temporary directory
                to write the data to. If not given, uses default value set
                at initialization.
            **update_meta_cfg: Can be used to update the meta configuration

        Returns:
            Multiverse: The created Multiverse object

        Raises:
            ValueError: If more than one of the run config selecting arguments
                (``from_cfg``, ``from_cfg_set``, ``run_cfg_path``) were given.
        """
        # A dict that can be filled with objects to store in self._mvs
        objs_to_store = dict()

        # May want to use config file relative to base directory or a run
        # directory from a config set. Need to check that only one of them
        # was given.
        if bool(from_cfg) + bool(from_cfg_set) + bool(run_cfg_path) > 1:
            raise ValueError(
                "Can pass at most one of the arguments `from_cfg`, "
                "`from_cfg_set`, or `run_cfg_path` but got more than one!"
            )

        if from_cfg:
            if os.path.isabs(from_cfg) and not self.base_dir:
                raise ValueError(
                    "Missing base_dir to handle relative path in "
                    "`from_cfg` argument."
                )

            run_cfg_path = os.path.join(self.base_dir, from_cfg)

        elif from_cfg_set:
            run_cfg_path = self.get_config_set(from_cfg_set)["run"]

        # Check whether a temporary directory is desired
        use_tmpdir = use_tmpdir if use_tmpdir is not None else self._use_tmpdir
        if use_tmpdir:
            tmpdir = self._create_tmpdir()
            objs_to_store["out_dir"] = tmpdir

            # Use update_meta_cfg to communicate it to the Multiverse
            if "paths" not in update_meta_cfg:
                update_meta_cfg["paths"] = dict(out_dir=tmpdir.name)
            else:
                update_meta_cfg["paths"]["out_dir"] = tmpdir.name

        # Also set the exit handling value, if not already set
        # TODO do this in a more elegant way
        _se = self._sim_errors
        if _se and "worker_manager" not in update_meta_cfg:
            update_meta_cfg["worker_manager"] = dict(nonzero_exit_handling=_se)

        elif (
            _se
            and "nonzero_exit_handling"
            not in update_meta_cfg["worker_manager"]
        ):
            update_meta_cfg["worker_manager"]["nonzero_exit_handling"] = _se

        # else: entry was already set; don't set it again

        # Create the Multiverse and store it, to not let it go out of scope
        mv = Multiverse(
            model_name=self.name,
            info_bundle=self.info_bundle,
            run_cfg_path=run_cfg_path,
            **update_meta_cfg,
        )
        self._store_mv(mv, **objs_to_store)

        return mv

    def create_frozen_mv(self, **fmv_kwargs) -> FrozenMultiverse:
        """Create a :py:class:`utopya.multiverse.FrozenMultiverse`, coupling it
        to a run directory.

        Use this method if you want to load an *existing* simulation run.

        Args:
            **fmv_kwargs: Passed on to
                :py:meth:`utopya.multiverse.FrozenMultiverse.__init__`
        """
        mv = FrozenMultiverse(model_name=self.name, **fmv_kwargs)
        self._store_mv(mv)

        return mv

    def create_run_load(
        self,
        *,
        from_cfg: str = None,
        run_cfg_path: str = None,
        from_cfg_set: str = None,
        use_tmpdir: bool = None,
        print_tree: bool = True,
        **update_meta_cfg,
    ) -> Tuple[Multiverse, DataManager]:
        """Chains the :py:meth:`.create_mv`, ``mv.run``, and
        ``mv.dm.load_from_cfg`` methods calls together and returns a
        ``(Multiverse, DataManager)`` tuple.

        Args:
            from_cfg (str, optional): The name of the config file (relative to
                the base directory) to be used.
            from_cfg_set (str, optional): Name of the config set to retrieve
                the run config from. Mutually exclusive with ``from_cfg`` and
                ``run_cfg_path``.
            run_cfg_path (str, optional): The path of the run config to use.
                Can not be passed if ``from_cfg`` or ``from_cfg_set`` arguments
                were given.
            use_tmpdir (bool, optional): Whether to use a temporary directory
                to write the data to. If not given, uses default value set
                at initialization.
            print_tree (bool, optional): Whether to print the loaded data tree
            **update_meta_cfg: Arguments passed to the create_mv function

        Returns:
            Tuple[Multiverse, DataManager]: The created Multiverse and the
                corresponding DataManager (with data already loaded).
        """
        mv = self.create_mv(
            from_cfg=from_cfg,
            from_cfg_set=from_cfg_set,
            run_cfg_path=run_cfg_path,
            use_tmpdir=use_tmpdir,
            **update_meta_cfg,
        )
        mv.run()
        mv.dm.load_from_cfg(print_tree=print_tree)
        return mv, mv.dm

    # Config set retrieval ....................................................

    def get_config_set(self, name: str = None) -> Dict[str, str]:
        """Returns a configuration set: a dict containing paths to run and/or
        eval configuration files. These are accessible via the keys ``run``
        and ``eval``.

        Config sets are retrieved from multiple locations:

            *  The ``cfgs`` directory in the model's source directory
            *  The user-specified lookup directories, specified in the utopya
               configuration as ``config_set_search_dirs``
            *  If ``name`` is an absolute or relative path, and a directory
               exists at the specified location, the parent directory is
               interpreted as a search path.

        This uses :py:meth:`~utopya.model.Model.get_config_sets` to retrieve
        all available configuration sets from the above paths and then selects
        the one with the given ``name``.
        Config sets that are found later overwrite those with the same name
        found in previous searches and log a warning message (which can be
        controlled with the ``warn`` argument); sets are *not* merged.

        For more information, see :ref:`config_sets`.

        Args:
            name (str, optional): The name of the config set to retrieve. This
                may also be a local path, which is looked up prior to the
                default search directories.
        """
        search_dirs = []

        # The name argument may be path-like, and we may want to include that
        # as well, but only if the directory actually exists.
        # In that case, the last path segment should be regarded as the name.
        _path = os.path.normpath(os.path.expanduser(name))
        if not os.path.isabs(_path):
            _path = os.path.abspath(_path)

        # However, we need to robustly identify the actual config set name from
        # the path in such a case.
        # Simplest assumption: check if such a directory actually exists, in
        # which case the user probably wanted to search it for a config set.
        # If there is a typo in the path, the whole directory will be searched
        # in the error message anyway ...
        if os.path.isdir(_path):
            name = os.path.basename(_path)
            search_dirs.append(os.path.dirname(_path))
        else:
            log.remark(
                "Given config set name was not interpretable as an "
                "additional search path (or the directory does not exist)."
            )

        # Append the default search directories, then start searching
        search_dirs += self.default_config_set_search_dirs

        for search_dir in search_dirs:
            cfg_sets = self._find_config_sets(search_dir, cfg_sets={})

            if name not in cfg_sets:
                continue

            cfg_set = cfg_sets[name]
            log.note(
                "Found config set named '%s' (contains: %s).",
                name,
                ", ".join(f"{k}.yml" for k in cfg_set.keys() if k != "dir"),
            )
            return cfg_set

        # else: did not find the config set.
        # Provide a useful error message, in which the local directory is
        # *always* searched, such that the user does not have to check manually
        # whether the directory exists but simply sees in this error message
        # whether there was a typo in `name`.
        search_dirs = [
            os.path.dirname(_path)
        ] + self.default_config_set_search_dirs
        _search_dirs = "\n".join(f"  - {s}" for s in search_dirs)

        with _adjusted_log_levels(("utopya.model", logging.WARNING)):
            _avail = self.get_config_sets(search_dirs=search_dirs, warn=False)

        raise ValueError(
            f"No config set with name '{name}' could be found in any of the "
            f"following search directories:\n{_search_dirs}\n\n"
            "Check that a subdirectory of the desired name exists at any of "
            "the above locations and contains a `run.yml` and/or `eval.yml` "
            "file.\n"
            "Available config sets:\n"
            f"{_make_columns(_avail) if _avail else '  (none available)'}"
        )

    def get_config_sets(
        self,
        *,
        search_dirs: List[str] = None,
        warn: bool = True,
        cfg_sets: dict = None,
    ) -> Dict[str, dict]:
        """Searches for all available configuration sets in the given search
        directories, aggregating them into one dict.

        The search is done in *reverse* order of the paths given in
        ``search_dirs``, i.e. starting from those directories with the lowest
        precedence. If configuration sets with the same name are encountered,
        warnings are emitted, but the one with higher precedence (appearing
        more towards the front of ``search_dirs``, i.e. the later-searched one)
        will take precedence.

        .. note::

            This will *not* merge configuration sets from different search
            directories, e.g. if one contained only an eval configuration and
            the other contained only a run configuration, a warning will be
            emitted but the entry from the later-searched directory will be
            used.

        Args:
            search_dirs (List[str], optional): The directories to search
                sequentially for config sets. If not given, will use the
                default config set search directories, see
                :py:attr:`~utopya.model.Model.default_config_set_search_dirs`.
            warn (bool, optional): Whether to warn (via log message), if the
                search yields a config set with a name that already existed.
            cfg_sets (dict, optional): If given, aggregate newly found config
                sets into this dict. Otherwise, start with an empty one.
        """
        if search_dirs is None:
            search_dirs = self.default_config_set_search_dirs

        cfg_sets = cfg_sets if cfg_sets is not None else dict()

        for search_dir in reversed(search_dirs):
            cfg_sets = self._find_config_sets(
                search_dir, cfg_sets=cfg_sets, warn=warn
            )
        return cfg_sets

    # Helpers .................................................................

    def _store_mv(self, mv: Multiverse, **kwargs) -> None:
        """Stores a created Multiverse object and all the kwargs in a dict"""
        self._mvs.append(dict(mv=mv, **kwargs))

    def _create_tmpdir(self) -> TemporaryDirectory:
        """Create a TemporaryDirectory"""
        return TemporaryDirectory(
            prefix=self.name, suffix=f"_mv{len(self._mvs)}"
        )

    def _find_config_sets(
        self, search_dir: str, *, cfg_sets: dict, warn: bool = True
    ) -> Dict[str, dict]:
        """Looks for config sets in the given directory and aggregates them
        into the given ``cfg_sets`` dict, warning if an entry already exists.

        Args:
            search_dir (str): The directory to search for configuration sets.
                Can be an absolute or relative path; ``~`` is expanded.
            cfg_sets (dict): The dict to populate with the results, each entry
                being one config set.
            warn (bool, optional): Whether to warn (via log message) if an
                entry already exists.
        """
        # Make absolute
        search_dir = os.path.expanduser(search_dir)
        if not os.path.isabs(search_dir):
            search_dir = os.path.abspath(search_dir)

        log.remark("Searching for config sets in:\n  %s", search_dir)

        if not os.path.isdir(search_dir):
            log.remark("No directory found at given search path.")
            return cfg_sets

        dn = 0
        for cs_name in os.listdir(search_dir):
            log.debug("Inspecting subdirectory '%s' ...", cs_name)

            search_subdir = os.path.join(search_dir, cs_name)
            found_cfgs = dict()

            # Run configuration
            run_cfg = os.path.join(search_subdir, "run.yml")
            if os.path.exists(run_cfg):
                found_cfgs["run"] = run_cfg

            # Eval configuration
            eval_cfg = os.path.join(search_subdir, "eval.yml")
            if os.path.exists(eval_cfg):
                found_cfgs["eval"] = eval_cfg

            # Only add entry, if configs were found
            if found_cfgs:
                dn += 1
                if (
                    warn
                    and cs_name in cfg_sets
                    and search_subdir != cfg_sets[cs_name]["dir"]
                ):
                    log.caution(
                        "A config set named '%s' was already found at:\n  %s\n"
                        "It will be overwritten with the one found at:\n  %s",
                        cs_name,
                        cfg_sets[cs_name]["dir"],
                        search_subdir,
                    )

                cfg_sets[cs_name] = dict(dir=search_subdir, **found_cfgs)

        log.remark("Found %d config set%s.", dn, "s" if dn != 1 else "")
        return cfg_sets



"""Sets up the utopya package, test dependencies, and command line scripts"""

from setuptools import find_packages, setup

# .. Dependency lists .........................................................

INSTALL_DEPS = [
    "numpy < 2.0",
    "scipy",
    "matplotlib",
    "seaborn",
    "ruamel.yaml",
    "coloredlogs",
    "colorama",
    "click",
    "pydantic >= 2.0",
    "python-git-info",
    #
    # first-party packages
    "yayaml>=0.1.1",
    "paramspace>=2.6.1",
    "dantro>=0.19.2",
]
# NOTE When adding a new dependency, make sure to denote it in the isort
#      configuration, see pyproject.toml.

# Dependencies for running tests and general development of utopya
TEST_DEPS = [
    "pytest",
    "pytest-cov",
    "pytest-order",
    "pre-commit",
]

# Dependencies for building the utopya documentation
DOC_DEPS = [
    "sphinx>=5.3,<6",
    "sphinx-book-theme",
    "sphinx-togglebutton",
    "ipython>=7.0",
    "myst-parser[linkify]",
    "sphinx-click",
    "pytest",
]

# Optional dependencies
OPTIONAL_DEPS = [
    "networkx",
    "pygraphviz",
]


# .. Description ..............................................................

DESCRIPTION = "A simulation management and evaluation framework"
LONG_DESCRIPTION = """
``utopya``: A simulation management and evaluation framework
============================================================

The ``utopya`` package provides a simulation management and evaluation
framework with the following feature set:

- **Run model simulations** in parallel and on cluster architectures

  - Conveniently perform parameter sweeps of arbitrary parameters with the help
    of the `paramspace <https://gitlab.com/blsqr/paramspace>`_ package.

- A **powerful CLI** to run and evaluate models, including interactive plotting
- Integrates the `dantro <https://gitlab.com/utopia-project/dantro>`_
  **data processing pipeline**:

  - Loads data into a hierarchical data tree, supplying a uniform interface
  - Gives access to a configuration-based **data transformation framework**,
    separating data preprocessing from visualization for increased generality
  - Easy extensibility of plot creators via model-specific plot implementations

- A **versatile configuration interface** for both simulation and evaluation:

  - Assembling multi-level model configurations, including several default
    levels
  - Assembling plot configurations with multiple inheritance, reducing
    redundant definitions

- Model, project, and framework registration and handling
- Managing data output directories
- Tools to simplify model test implementations or working without a CLI
- ... and more

The ``utopya`` package evolved as part of the
`Utopia Project <https://utopia-project.org>`_ and provides the frontend of
the `Utopia modelling framework <https://gitlab.com/utopia-project/utopia>`_.
Having been outsourced from that project, it can be used with arbitrary model
implementations with a very low barrier for entry: in the simplest case, only
the path to an executable is required to run simulations.
With more compliance to the utopya interface, more features become available.

The ``utopya`` package is **open source software** released under the
`LGPLv3+ <https://www.gnu.org/licenses/lgpl-3.0.html>`_ license.

Visit the `utopya project website <https://gitlab.com/utopia-project/utopya>`_
or the `documentation page <https://utopya.readthedocs.io/>`_ for more
information about utopya.
"""


# .............................................................................


def find_version(*file_paths) -> str:
    """Tries to extract a version from the given path sequence"""
    import codecs
    import os
    import re

    def read(*parts):
        """Reads a file from the given path sequence, relative to this file"""
        here = os.path.abspath(os.path.dirname(__file__))
        with codecs.open(os.path.join(here, *parts), "r") as fp:
            return fp.read()

    # Read the file and match the __version__ string
    file = read(*file_paths)
    match = re.search(r"^__version__\s?=\s?['\"]([^'\"]*)['\"]", file, re.M)
    if match:
        return match.group(1)
    raise RuntimeError("Unable to find version string in " + str(file_paths))


# .............................................................................


setup(
    name="utopya",
    #
    # Package information
    version=find_version("utopya", "__init__.py"),
    #
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    url="https://gitlab.com/utopia-project/utopya",
    author="utopya developers",
    author_email=("Yunus Sevinchan <yunus.sevinchan@hu-berlin.de>"),
    classifiers=[
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        #
        "License :: OSI Approved :: GNU Lesser General Public License v3 or later (LGPLv3+)",
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        #
        "Topic :: Utilities",
        "Topic :: Scientific/Engineering",
    ],
    #
    # Package content
    packages=find_packages(exclude=("demo", "tests")),
    package_data=dict(utopya=["cfg/*.yml"]),
    data_files=[
        ("", ["COPYING", "COPYING.LESSER", "README.md", "CHANGELOG.md"]),
    ],
    #
    # Dependencies
    install_requires=INSTALL_DEPS,
    extras_require=dict(
        test=TEST_DEPS,
        doc=DOC_DEPS,
        opt=OPTIONAL_DEPS,
        dev=TEST_DEPS + DOC_DEPS + OPTIONAL_DEPS,
    ),
    #
    # Command line scripts
    entry_points={
        "console_scripts": [
            "utopya = utopya_cli.cli:cli",
            "utopia = utopya_cli.cli:cli",
        ],
    },
)



"""Implements functionality to copy relevant model files"""

import glob
import logging
import os
from typing import Callable, Dict, List, Sequence, Tuple

import click

log = logging.getLogger(__name__)

FILE_EXTENSIONS = {
    "c": (".c", ".h"),
    "c++": (".c", ".h", ".cc", ".hh", ".cpp"),
    "python": (".py", ".pyx"),
    "yaml": (".yml", ".yaml"),
}
"""A map of language-specific file extensions that are used for file search"""

LANGUAGE_ALIASES = {
    "cpp": "c++",
    "py": "python",
    "py2": "python",
    "py3": "python",
}
"""A map of language specifier aliases that can be used to find a normalized
language specifier"""

# -----------------------------------------------------------------------------


def abbrev_path(p: str) -> str:
    """Abbreviates a path by reversing expanduser and making a path absolute"""
    p = os.path.abspath(p)
    homedir = os.path.expanduser("~")
    if p.startswith(homedir):
        p = "~" + p[len(homedir) :]
    return p


def apply_replacements(
    s: str, *replacements: Sequence[Tuple[str, str]]
) -> str:
    """Applies multiple replacements onto the given string"""
    for replacement in replacements:
        s = s.replace(*replacement)
    return s


def create_file_map(
    *,
    source_dir: str,
    target_dir: str,
    abs_file_map: dict,
    replacements: Sequence[Tuple[str, str]],
    skip_exts: Sequence[str] = None,
    glob_args: Sequence[str] = ("**",),
) -> dict:
    """Given a file list with absolute paths, aggregates the file path
    changes into ``abs_file_map`` and gathers the relative file path
    changes into the returned dict.

    The file name is changed according to the specified replacements.

    Args:
        source_dir (str): The source directory to look for files in using
            glob and the ``glob_args``. Note that directories are not
            matched.
        target_dir (str): The target directory of the renamed files
        abs_file_map (dict): The mutable file map that the absolute file
            path changes are aggregated in.
        replacements (Sequence[Tuple[str, str]]): The replacement
            specifications, applied to the relative paths.
        glob_args (Sequence[str], optional): The glob arguments to match
            files within the source directory. By default, this matches
            all files, also down the source directory tree. The glob
            ``recursive`` option is enabled.

    Returns:
        dict: The file map relative to source and target dir.
    """
    files = glob.glob(os.path.join(source_dir, *glob_args), recursive=True)
    rel_file_map = dict()

    for fpath in files:
        if os.path.isdir(fpath) or not os.path.exists(fpath):
            continue

        if skip_exts and os.path.splitext(fpath)[1] in skip_exts:
            continue

        rel_fpath = os.path.relpath(fpath, start=source_dir)
        new_rel_fpath = apply_replacements(rel_fpath, *replacements)

        abs_file_map[fpath] = os.path.join(target_dir, new_rel_fpath)
        rel_file_map[rel_fpath] = new_rel_fpath

    return rel_file_map


def show_file_map(
    *,
    file_map: dict,
    source_dir: str,
    target_dir: str,
    desc: str,
    _log: logging.Logger = log,
):
    """Shows a human-readable version of the given (relative) file map
    which copies from the source directory tree to the target directory
    tree.
    """
    _log.progress(f"\n--- Found {len(file_map):d} {desc:s} files.")
    _log.note(f"Source directory:\n    {abbrev_path(source_dir)}\n")
    _log.note(f"Target directory:\n    {abbrev_path(target_dir)}\n")

    if not file_map:
        _log.remark(
            "No files found; check the file extensions and paths that are "
            "to be included.\n"
        )
        return

    max_key_len = min(max(len(k) for k in file_map), 32)
    files = "\n".join(
        "    {:{l:d}s}  ->  {:s}".format(k, v, l=max_key_len)
        for k, v in file_map.items()
    )

    _log.note(f"File mapping:\n{files}")


# .............................................................................


def add_model_to_cmakelists(
    *, fpath: str, new_name: str, dry_run: bool, _log: logging.Logger = log
):
    """Adds the relevant add_subdirectory command to the CMakeLists file
    at the specified path.

    Assumes an ascending alphabetical list of ``add_subdirectory`` commands
    and adds the new command at a suitable place.
    If no ``add_subdirectory`` command is found, inserts at the end.

    Args:
        fpath (str): The absolute path of the CMakeLists.txt file
        new_name (str): The new model name to add to it
        dry_run (bool): If true, will not write but show a preview instead
    """
    if not os.path.exists(fpath):
        _log.warning(
            f"No CMakeLists.txt file found in expected location:\n  {fpath}"
        )
        _log.remark("If you want CMake postprocessing, add a file there.")
        return

    # Read the file
    with open(fpath) as f:
        lines = f.readlines()

    # Find the line to add the add_subdirectory command at
    insert_idx = None
    i = 0
    for i, line in enumerate(lines):
        if not line.startswith("add_subdirectory"):
            continue

        insert_idx = i
        _model = line[len("add_subdirectory(") : -2]
        if _model.lower() > new_name.lower():
            break
    else:
        # Did not break
        if insert_idx is None:
            # No add_subdirectory found, insert at end of file
            _log.warning(
                "Found no add_subdirectory commands and thus do not know "
                "where to insert the add_subdirectory command for the copied "
                "model's subdirectory; inserting at the end of the file ..."
            )
            insert_idx = i
        else:
            # Insert behind the last add_subdirectory command
            insert_idx += 1

    lines.insert(insert_idx, f"add_subdirectory({new_name})\n")

    if not dry_run:
        _log.remark("Now writing ...")
        with open(fpath, "w") as f:
            f.writelines(lines)

        _log.progress(
            f"add_subdirectory('{new_name}') added to\n"
            f"   {abbrev_path(fpath)}"
        )

    else:
        _log.note(
            "This is a dry run.\nThe following is a preview of how the new\n"
            f"   {abbrev_path(fpath)}\n"
            "file _would_ look like:\n"
        )
        _log.info("-" * 79 + "\n")
        _log.info("".join(lines))
        _log.info("-" * 79)


def postprocess_copied_model(
    *,
    new_name: str,
    info_bundle,
    target_project,
    dry_run: bool,
    file_map: str,
    impl_target_dir: str,
    _log: logging.Logger = log,
    # TODO Should pass information in a more general form here
    #
    # User-controlled parameters
    enabled: bool = True,
    cmake: dict = None,
):
    """Performs postprocessing routines, e.g. file adaptations that embed the
    copied model into the build system."""
    _log.hilight("\nPostprocessing copied model ...\n")
    if not enabled:
        _log.remark("Post-processing routines were disabled.")
        return

    # .. CMake-related stuff ..................................................
    cmake = cmake if cmake else {}
    cmake_enabled = cmake.get("enabled", "auto")
    if cmake_enabled == "auto":
        if any(p.endswith("CMakeLists.txt") for p in file_map.values()):
            cmake_enabled = True
            _log.note(
                "Found CMakeLists.txt file in file map, thus auto-enabling "
                "CMake postprocessing routine."
            )
            _log.remark(
                "Set postprocessing.cmake.enabled to False to suppress this."
            )
        else:
            cmake_enabled = False

    if cmake_enabled:
        cmakelists_fpath = os.path.abspath(
            os.path.join(impl_target_dir, "../CMakeLists.txt")
        )

        _log.progress(
            "Inserting add_subdirectory command for model directory ..."
        )
        add_model_to_cmakelists(
            fpath=cmakelists_fpath,
            new_name=new_name,
            dry_run=dry_run,
            _log=_log,
        )

    _log.success("\nFinished post-processing the copied model.")


# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------


def copy_model_files(
    *,
    model_name: str,
    label: str = None,
    new_name: str = None,
    target_project: str = None,
    skip_exts: Sequence[str] = None,
    dry_run: bool = False,
    prompts_confirmed: bool = False,
    raise_exc: bool = False,
    postprocess: dict = None,
    _log: logging.Logger = log,
) -> None:
    """A helper function to conveniently copy model-related files, rename them,
    and adjust their content to the new name as well.

    Args:
        model_name (str): The name of the model to copy
        new_name (str, optional): The new name of the model. This may not
            conflict with any already existing model name in the model
            registry.
        target_project (str, optional): The name of the project to copy the
            model to. It needs to be a registered Utopia project.
        add_to_cmakelists (bool, optional): Whether to add the new model to the
            corresponding CMakeLists.txt file.
        dry_run (bool, optional): If given, no write or copy operations will be
            carried out.

    Raises:
        ValueError: Upon bad arguments
    """
    from utopya import MODELS, PROJECTS
    from utopya.model_registry import get_info_bundle

    _indent = " " * 3

    def handle_exc(exc: Exception, desc: str):
        """A little exception handler function for read/write errors"""
        _log.error(f"{_indent}{desc} failed with {type(exc).__name__}: {exc}.")
        if raise_exc:
            _log.remark("Unset --debug flag to ignore error and continue.")
            raise

        _log.remark(
            f"{_indent}If you want this file copied and refactored, "
            "you will have to do it manually."
        )
        _log.remark("Set --debug flag to raise instead of continuing.")

    # Gather information on model, project, and replacements . . . . . . . . .
    # Get the model information
    info_bundle = get_info_bundle(model_name=model_name, bundle_label=label)
    _log.info(
        f"\nModel selected to copy:      {info_bundle.model_name}  "
        f"(from project: {info_bundle.project_name})"
    )

    # Check if the name is not already taken, being case-insensitive
    _log.info(f"Name of the new model:       {new_name}")

    if new_name.lower() in [n.lower() for n in MODELS.keys()]:
        _avail = ", ".join(MODELS.keys())
        raise ValueError(
            f"A model with name '{new_name}' is already registered! "
            "Make sure that the name is unique. If you keep receiving this "
            "error despite no other model with this name being implemented, "
            "remove the entry from the model registry, e.g. via the "
            "`utopya models rm` CLI command.\n"
            f"Already registered models: {_avail}"
        )

    # Find out about the project that the files are copied _to_
    _log.info(f"Project to copy to:          {target_project}")
    target_project = PROJECTS[target_project]

    # Determine replacements
    replacements = [
        (model_name, new_name),
        (model_name.lower(), new_name.lower()),
        (model_name.upper(), new_name.upper()),
    ]

    # Generate the file maps . . . . . . . . . . . . . . . . . . . . . . . . .
    _log.hilight("\nSearching for files to copy ...")

    # Prepare file extension skipping
    skip_exts = [e if e.startswith(".") else f".{e}" for e in skip_exts]
    _log.remark(
        f"File extensions that are skipped:\n    {' '.join(skip_exts)}\n"
    )

    # The mapping of all files that are to be copied and in which the content
    # is to be replaced. It maps absolute source file paths to absolute target
    # file paths.
    file_map = dict()

    # Relative file maps, created below
    impl_file_map = dict()
    py_t_file_map = dict()
    py_p_file_map = dict()

    # Find out the target directories
    target_models_dir = target_project.paths.get("models_dir")
    target_py_t_dir = target_project.paths.get("py_tests_dir")
    target_py_p_dir = target_project.paths.get("py_plots_dir")

    # Define the source and target directory paths of the implementation and
    # the python-related files, if the path information is available.
    impl_source_dir = info_bundle.paths["source_dir"]
    impl_target_dir = os.path.join(target_models_dir, new_name)
    impl_file_map = create_file_map(
        source_dir=impl_source_dir,
        target_dir=impl_target_dir,
        abs_file_map=file_map,
        replacements=replacements,
        skip_exts=skip_exts,
    )

    if target_py_t_dir and info_bundle.paths.get("py_tests_dir"):
        py_t_source_dir = info_bundle.paths["py_tests_dir"]
        py_t_target_dir = os.path.join(target_py_t_dir, new_name)
        py_t_file_map = create_file_map(
            source_dir=py_t_source_dir,
            target_dir=py_t_target_dir,
            abs_file_map=file_map,
            replacements=replacements,
            skip_exts=skip_exts,
        )
    else:
        _log.note(
            f"Model '{model_name}' and/or target project "
            f"'{target_project.name}' do not define a 'py_tests_dir' "
            "directory."
        )

    if target_py_p_dir and info_bundle.paths.get("py_plots_dir"):
        py_p_source_dir = info_bundle.paths["py_plots_dir"]
        py_p_target_dir = os.path.join(target_py_p_dir, new_name)
        py_p_file_map = create_file_map(
            source_dir=py_p_source_dir,
            target_dir=py_p_target_dir,
            abs_file_map=file_map,
            replacements=replacements,
            skip_exts=skip_exts,
        )
    else:
        _log.note(
            f"Model '{model_name}' and/or target project "
            f"'{target_project.name}' do not define a 'py_plots_dir' "
            "directory."
        )

    # Gathered all information now. . . . . . . . . . . . . . . . . . . . . . .
    # Inform about the file changes and the replacement in them.
    show_file_map(
        file_map=impl_file_map,
        desc="model implementation",
        source_dir=impl_source_dir,
        target_dir=impl_target_dir,
        _log=_log,
    )

    if py_t_file_map:
        show_file_map(
            file_map=py_t_file_map,
            desc="python model test",
            source_dir=py_t_source_dir,
            target_dir=py_t_target_dir,
            _log=_log,
        )

    if py_p_file_map:
        show_file_map(
            file_map=py_p_file_map,
            desc="python model plot",
            source_dir=py_p_source_dir,
            target_dir=py_p_target_dir,
            _log=_log,
        )

    if not file_map:
        _log.caution("No files found to copy! Not continuing ...")
        return

    max_repl_len = max(len(rs) for rs, _ in replacements)
    repl_info = [
        "    '{:{l:d}s}'  ->  '{:s}'".format(*repl, l=max_repl_len)
        for repl in replacements
    ]
    _repl_info = "\n".join(repl_info)
    _log.info(
        f"\nInside all of these {len(file_map):d} files, the following string "
        "replacements will be carried out:"
    )
    _log.note(f"{_repl_info}\n")

    # Inform about further procedure and prompt confirmation
    _log.progress("Ready for copying now.\n")
    if dry_run:
        _log.hilight("--- THIS IS A DRY RUN. ---")
        _log.remark("Copy and write operations below are NOT operational.\n")
    else:
        _log.caution("This is not a drill.")
        _log.remark(
            "If you want to see the effect of copying first, "
            "consider making a dry-run by setting the --dry-run CLI flag.\n"
        )

    if not prompts_confirmed and not click.confirm("Proceed?"):
        _log.caution("Not proceeding.")
        return

    _log.hilight("\nCopying and refactoring ...")

    # Now, the actual copying . . . . . . . . . . . . . . . . . . . . . . . . .
    for i, (src_fpath, target_fpath) in enumerate(file_map.items()):
        _log.progress(f"\nFile {i + 1:d}/{len(file_map):d} ...")
        _log.remark(
            f"{_indent}   {abbrev_path(src_fpath):s}\n"
            f"{_indent}-> {abbrev_path(target_fpath):s}"
        )

        try:
            with open(src_fpath) as src_file:
                src_lines = src_file.read()

        except Exception as exc:
            handle_exc(exc, "Reading")
            continue

        target_lines = apply_replacements(src_lines, *replacements)

        if dry_run:
            continue

        # Create directories and write the file; failing if it already exists
        os.makedirs(os.path.dirname(target_fpath), exist_ok=True)
        try:
            with open(target_fpath, mode="x") as target_file:
                target_file.write(target_lines)

        except Exception as exc:
            handle_exc(exc, "Copying")
            continue

    _log.success("\nFinished copying and refactoring.")
    if dry_run:
        _log.remark("Reminder: This is a dry run. No copying happened.")

    # Invoke postprocessing, which may also be disabled
    postprocess_copied_model(
        new_name=new_name,
        info_bundle=info_bundle,
        target_project=target_project,
        file_map=file_map,
        impl_target_dir=impl_target_dir,
        dry_run=dry_run,
        _log=_log,
        **(postprocess if postprocess else {}),
    )

    if dry_run:
        _log.hilight(f"\nDry-run for copying '{model_name}' model finished.")
        _log.remark("To actually copy files, unset the --dry-run flag.")
    else:
        _log.hilight(f"\nSuccessfully copied the '{model_name}' model.")
        _log.remark("Hint: Use `utopya models register` to register it.")
